from __future__ import annotations

from bisect import bisect_right
from collections.abc import Hashable, Iterator
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import dataclass
from datetime import timedelta
from functools import cache
from logging import getLogger
from math import floor
from typing import Any, Literal, NamedTuple, Self

from jetpytools import clamp, complex_hash, cround
from PySide6.QtCore import QEvent, QLineF, QPoint, QPointF, QRectF, QSignalBlocker, QSize, Qt, QTime, Signal
from PySide6.QtGui import (
    QColor,
    QContextMenuEvent,
    QCursor,
    QFontMetrics,
    QIcon,
    QMouseEvent,
    QMoveEvent,
    QPainter,
    QPaintEvent,
    QPalette,
    QPen,
    QResizeEvent,
    QRgba64,
    QValidator,
)
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QMenu,
    QSizePolicy,
    QSlider,
    QSpinBox,
    QTimeEdit,
    QToolButton,
    QToolTip,
    QWidget,
    QWidgetAction,
)
from vsengine.loops import get_loop

from ...assets import IconName, IconReloadMixin
from ...vsenv import run_in_loop
from ..outputs import AudioOutput
from ..settings import SettingsManager
from .components import SegmentedControl

logger = getLogger(__name__)


class Frame(int):
    """Frame number type."""


class Time(timedelta):
    """Time type."""

    def to_qtime(self) -> QTime:
        """Convert a Time object to a QTime object."""
        # QTime expects milliseconds since the start of the day
        total_ms = cround(self.total_seconds() * 1000)

        # Caps at 23:59:59.999. If delta > 24h, it wraps around.
        return QTime.fromMSecsSinceStartOfDay(total_ms)

    def to_ts(self, fmt: str = "{H:02d}:{M:02d}:{S:02d}.{ms:03d}") -> str:
        """
        Formats a timedelta object using standard Python formatting syntax.

        Available keys:
        {D}  : Days
        {H}  : Hours (0-23)
        {M}  : Minutes (0-59)
        {S}  : Seconds (0-59)
        {ms} : Milliseconds (0-999)
        {us} : Microseconds (0-999999)

        Total duration keys:
        {th} : Total Hours (e.g., 100 hours)
        {tm} : Total Minutes
        {ts} : Total Seconds

        Example:
            ```python
            # 1. Standard Clock format (Padding with :02d)
            # Output: "26:05:03"
            print(time.to_ts(td, "{th:02d}:{M:02d}:{S:02d}"))

            # 2. Detailed format
            # Output: "1 days, 02 hours, 05 minutes"
            print(time.to_ts(td, "{D} days, {H:02d} hours, {M:02d} minutes"))

            # 3. With Milliseconds
            # Output: "02:05:03.500"
            print(time.to_ts(td, "{H:02d}:{M:02d}:{S:02d}.{ms:03d}"))
            ```

        """
        total_seconds = int(self.total_seconds())

        days = self.days
        hours, remainder = divmod(self.seconds, 3600)
        minutes, seconds = divmod(remainder, 60)

        milliseconds = cround(self.microseconds / 1000)

        format_data = {
            "D": days,
            "H": hours,
            "M": minutes,
            "S": seconds,
            "ms": milliseconds,
            "us": self.microseconds,
            # Total durations (useful for "26 hours ago")
            "th": total_seconds // 3600,
            "tm": total_seconds // 60,
            "ts": total_seconds,
        }

        return fmt.format(**format_data)

    @classmethod
    def from_qtime(cls, qtime: QTime) -> Self:
        """Convert a QTime object to a Time object."""
        return cls(milliseconds=qtime.msecsSinceStartOfDay())


@cache
def generate_label_format(notch_interval_t: Time, end_time: Time) -> str:
    if end_time >= Time(hours=1):
        return "{H}:{M:02d}:{S:02d}"

    if notch_interval_t >= Time(minutes=1):
        return "{M}:{S:02d}"

    if end_time > Time(seconds=10):
        return "{M}:{S:02d}"

    return "{S}.{ms:03d}"


class Notch[T: (Time, Frame)]:
    """Represents a notch marker on the timeline."""

    class CacheKey(NamedTuple):
        rect: QRectF
        total_frames: int

    class CacheValue[T0: (Time, Frame)](NamedTuple):
        scroll_rect: QRectF
        labels_notches: list[Notch[T0]]
        rects_to_draw: list[tuple[QRectF, str]]

    class CacheEntry[T1: (Time, Frame)](NamedTuple):
        key: Notch.CacheKey
        value: Notch.CacheValue[T1]

    def __init__(
        self,
        data: T,
        end_data: T | None = None,
        color: Qt.GlobalColor | QColor | QRgba64 | str | int | None = None,
        line: QLineF | None = None,
        end_line: QLineF | None = None,
        label: str = "",
    ) -> None:
        self.data: T = data
        self.end_data: T | None = end_data
        self.color = QColor(color) if color is not None else QColor(Qt.GlobalColor.black)
        self.line = line if line is not None else QLineF()
        self.end_line = end_line
        self.label = label

    def draw(self, painter: QPainter, scroll_rect: QRectF, range_alpha: int = 80, cosmetic: bool = False) -> None:
        pen = QPen(self.color, 1)

        if cosmetic:
            pen.setCosmetic(True)

        painter.setPen(pen)

        if self.end_line is not None:
            x1, x2 = self.line.x1(), self.end_line.x1()
            fill_color = QColor(self.color)
            fill_color.setAlpha(range_alpha)
            painter.fillRect(QRectF(min(x1, x2), scroll_rect.top(), abs(x2 - x1), scroll_rect.height()), fill_color)
            painter.drawLine(self.line)
            painter.drawLine(self.end_line)
        else:
            painter.drawLine(self.line)


class CustomNotch[T: (Time, Frame)](Notch[T]):
    def __init__(
        self,
        id: Hashable,
        data: T,
        end_data: T | None = None,
        color: Qt.GlobalColor | QColor | QRgba64 | str | int | None = None,
        line: QLineF | None = None,
        end_line: QLineF | None = None,
        label: str = "",
    ) -> None:
        # mypy bug
        super().__init__(data, end_data, color, line, end_line, label)  # type: ignore[arg-type]
        self.id = id

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        return self.id == other.id if isinstance(other, CustomNotch) else NotImplemented


class HoverLabel(NamedTuple):
    text: str
    x: float
    color: QColor
    y_offset: float = 0.0


class TimelineHoverPopup(QWidget):
    # Constants
    NUM_LAYERS = 3
    POPUP_CORNER_RADIUS = 5
    PILL_CORNER_RADIUS = 4
    LABEL_PADDING_H = 4
    LABEL_PADDING_V = 2
    LABEL_STAGGER_OFFSETS = (0, -20, -40)
    LABEL_SPACING = 10
    ZOOMED_NOTCH_SPACING = 75
    RANGE_FILL_ALPHA = 80

    def __init__(self, parent: Timeline) -> None:
        super().__init__(None, Qt.WindowType.ToolTip | Qt.WindowType.FramelessWindowHint)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating)
        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)

        self._timeline = parent
        self.radius = 50
        self.zoom_factor = 4.0
        self.hover_x = -1

    def paintEvent(self, event: QPaintEvent) -> None:
        if self.hover_x < 0:
            return

        popup_rect = QRectF(self.rect())

        with QPainter(self) as painter:
            self._draw_background(painter, popup_rect)

            # Setup clipping and base transform
            painter.setClipRect(popup_rect)
            y_offset = self.height() - self._timeline.height()
            painter.translate(self.radius, y_offset)
            painter.scale(self.zoom_factor, 1.0)
            painter.translate(-self.hover_x, 0)

            self._draw_zoomed_notches(painter)
            self._draw_scroll_bar(painter)
            self._draw_custom_notches(painter, popup_rect, y_offset)
            self._draw_indicators(painter)

    def update_state(self, hover_x: int) -> None:
        self.zoom_factor = SettingsManager.global_settings.timeline.hover_zoom_factor
        self.radius = SettingsManager.global_settings.timeline.hover_zoom_radius
        self.hover_x = hover_x

        popup_width = self.radius * 2
        popup_height = round(self._timeline.height() * (1 + 0.33 * self.NUM_LAYERS))

        global_pos = self._timeline.mapToGlobal(QPoint(hover_x, 0))

        # Position popup above the timeline
        x = global_pos.x() - self.radius
        y = global_pos.y() - popup_height - 10

        self.setGeometry(x, y, popup_width, popup_height)
        self.update()

    def _draw_background(self, painter: QPainter, popup_rect: QRectF) -> None:
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        painter.fillRect(popup_rect, self._timeline.palette().color(self._timeline.BACKGROUND_COLOR))
        painter.setPen(QPen(self._timeline.palette().color(self._timeline.TEXT_COLOR), 1))
        painter.drawRoundedRect(popup_rect, self.POPUP_CORNER_RADIUS, self.POPUP_CORNER_RADIUS)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, on=False)

    def _draw_zoomed_notches(self, painter: QPainter) -> None:
        # Generate and draw main timeline notches for the zoomed view.
        fm = painter.fontMetrics()

        # Calculate viewport in timeline X coordinates
        half_view_x = self.radius / self.zoom_factor
        view_start_x = self.hover_x - half_view_x
        view_end_x = self.hover_x + half_view_x

        # Target interval in original space
        target_interval_x = self.ZOOMED_NOTCH_SPACING / self.zoom_factor

        tmp_notches = list[tuple[int, str]]()
        if self._timeline.mode == "frame":
            interval_f = self._timeline.calculate_notch_interval_f(int(target_interval_x))
            # Improved accuracy: clamp and extend by 1 to avoid rounding issues
            start_f = max(0, self._timeline.x_to_frame(int(view_start_x)) - 1)
            curr_f = Frame((start_f // interval_f) * interval_f)

            while curr_f < self._timeline.total_frames:
                x = self._timeline.cursor_to_x(curr_f)
                if x > view_end_x + target_interval_x:
                    break

                if x >= view_start_x - target_interval_x:
                    tmp_notches.append((x, str(curr_f)))

                curr_f = Frame(curr_f + interval_f)
        else:
            interval_t = self._timeline.calculate_notch_interval_t(int(target_interval_x))
            start_t = self._timeline.x_to_time(max(0, int(view_start_x)))
            interval_secs = interval_t.total_seconds()

            if interval_secs > 0:
                curr_secs = max(0, (start_t.total_seconds() // interval_secs) * interval_secs)
                curr_t = Time(seconds=curr_secs)
                label_format = generate_label_format(interval_t, self._timeline.total_time)

                while curr_t <= self._timeline.total_time:
                    x = self._timeline.cursor_to_x(curr_t)
                    if x > view_end_x + target_interval_x:
                        break

                    if x >= view_start_x - target_interval_x:
                        tmp_notches.append((x, curr_t.to_ts(label_format)))

                    curr_t = Time(seconds=curr_t.total_seconds() + interval_secs)

        # Draw notches and labels
        notch_pen = QPen(self._timeline.palette().color(self._timeline.TEXT_COLOR), 1)
        notch_pen.setCosmetic(True)
        painter.setPen(notch_pen)

        lnotch_y = self._timeline.rect_f.top() + self._timeline.font_height + self._timeline.notch_height + 5
        lnotch_top = lnotch_y - self._timeline.notch_height

        for nx, ntext in tmp_notches:
            painter.drawLine(QLineF(nx, lnotch_y, nx, lnotch_top))

            # Map to zoomed space for 1:1 text
            zoomed_pos = painter.transform().map(QPointF(nx, lnotch_top))

            painter.save()
            painter.resetTransform()
            painter.setPen(self._timeline.palette().color(self._timeline.TEXT_COLOR))
            text_width = fm.horizontalAdvance(ntext)
            painter.drawText(QPointF(zoomed_pos.x() - text_width / 2, zoomed_pos.y() - fm.descent()), ntext)
            painter.restore()

    def _draw_scroll_bar(self, painter: QPainter) -> None:
        painter.fillRect(self._timeline.scroll_rect, self._timeline.palette().color(self._timeline.SCROLL_BAR_COLOR))

    def _draw_custom_notches(self, painter: QPainter, popup_rect: QRectF, y_offset: float) -> None:
        # Draw provider notches (bookmarks, keyframes) with staggered labels.
        fm = painter.fontMetrics()

        # Collect labels with range support
        all_labels = list[HoverLabel]()
        for provider_notches in self._timeline.custom_notches.values():
            for p_notch in provider_notches:
                if not (label_text := self._format_notch_label(p_notch)):
                    continue

                x_pos = painter.transform().map(QPointF(p_notch.line.x1(), 0)).x()
                all_labels.append(HoverLabel(text=label_text, x=x_pos, color=p_notch.color))

        # Sort and stagger
        all_labels = sorted(all_labels, key=lambda lbl: lbl.x)
        staggered_labels = self._apply_staggering(all_labels, fm)

        # Draw labels with leader lines and pills
        for lbl in staggered_labels:
            self._draw_label_pill(painter, lbl, popup_rect, y_offset, fm)

        # Draw notches and range fills in scaled space
        for provider_notches in self._timeline.custom_notches.values():
            for p_notch in provider_notches:
                p_notch.draw(painter, self._timeline.scroll_rect, self.RANGE_FILL_ALPHA, cosmetic=True)

    def _format_notch_label(self, notch: Notch[Any]) -> str:
        base_label = notch.label

        if notch.end_data is not None:
            # Format: [start, end] label
            start_str = self._format_data_value(notch.data)
            end_str = self._format_data_value(notch.end_data)
            range_prefix = f"[{start_str}, {end_str}]"
            return f"{range_prefix} {base_label}" if base_label else range_prefix

            # Handle format strings in label
        return base_label.format(self._format_data_value(notch.data)) if "{" in base_label else base_label

    def _format_data_value(self, data: Frame | Time) -> str:
        match data, self._timeline.mode:
            case Time(), "time":
                return data.to_ts("{H:02d}:{M:02d}:{S:02d}.{ms:03d}")
            case Time(), "frame":
                # In frame mode, convert time to frame for display
                return str(self._timeline.x_to_frame(self._timeline.cursor_to_x(data)))
            case _:
                return str(data)

    def _apply_staggering(self, labels: list[HoverLabel], fm: QFontMetrics) -> list[HoverLabel]:
        if not labels:
            return []

        # Apply vertical staggering to overlapping labels.

        staggered = list[HoverLabel]()
        level_index = 0
        last_x = -9999.0

        for lbl in labels:
            if lbl.x - last_x < fm.horizontalAdvance(lbl.text) + self.LABEL_SPACING:
                level_index = (level_index + 1) % len(self.LABEL_STAGGER_OFFSETS)
            else:
                level_index = 0

            staggered.append(lbl._replace(y_offset=self.LABEL_STAGGER_OFFSETS[level_index]))
            last_x = lbl.x

        return staggered

    def _draw_label_pill(
        self,
        painter: QPainter,
        lbl: HoverLabel,
        popup_rect: QRectF,
        y_offset: float,
        fm: QFontMetrics,
    ) -> None:
        # Draw a single label with leader line and pill background.
        text_width = fm.horizontalAdvance(lbl.text)
        text_height = fm.height()

        painter.save()
        painter.resetTransform()

        p_y = self._timeline.scroll_rect.top() + y_offset - text_height - 5 + lbl.y_offset
        t_rect = QRectF(
            lbl.x - text_width / 2 - self.LABEL_PADDING_H,
            p_y - self.LABEL_PADDING_V,
            text_width + self.LABEL_PADDING_H * 2,
            text_height + self.LABEL_PADDING_V * 2,
        )

        if popup_rect.intersects(t_rect):
            # Leader line
            leader_pen = QPen(self._timeline.palette().color(self._timeline.TEXT_COLOR), 1)
            leader_pen.setCosmetic(True)
            painter.setPen(leader_pen)
            painter.setOpacity(0.5)
            painter.drawLine(QPointF(lbl.x, self._timeline.scroll_rect.top()), QPointF(lbl.x, t_rect.bottom()))
            painter.setOpacity(1.0)

            # Pill background
            painter.setRenderHint(QPainter.RenderHint.Antialiasing)
            bg_color = self._timeline.palette().color(self._timeline.BACKGROUND_COLOR)
            painter.setBrush(bg_color)
            painter.setPen(QPen(lbl.color, 1))
            painter.drawRoundedRect(t_rect, self.PILL_CORNER_RADIUS, self.PILL_CORNER_RADIUS)

            # Label text
            painter.setPen(self._timeline.palette().color(self._timeline.TEXT_COLOR))
            painter.drawText(t_rect, Qt.AlignmentFlag.AlignCenter, lbl.text)

        painter.restore()

    def _draw_indicators(self, painter: QPainter) -> None:
        # Draw hover dash and playback cursor.
        hover_pen = QPen(
            self._timeline.palette().color(self._timeline.BACKGROUND_COLOR),
            1,
            Qt.PenStyle.DashLine,
        )
        hover_pen.setCosmetic(True)
        painter.setPen(hover_pen)
        painter.drawLine(
            QLineF(
                self.hover_x,
                self._timeline.scroll_rect.top(),
                self.hover_x,
                self._timeline.scroll_rect.bottom(),
            )
        )

        cursor_pen = QPen(Qt.GlobalColor.black, 2)
        cursor_pen.setCosmetic(True)
        painter.setPen(cursor_pen)
        painter.drawLine(
            QLineF(
                self._timeline.cursor_x,
                self._timeline.scroll_rect.top(),
                self._timeline.cursor_x,
                self._timeline.scroll_rect.bottom(),
            )
        )


class Timeline(QWidget):
    # Signal emits (Frame, Time) when the user clicks on the timeline
    clicked = Signal(object, object)

    # Predefined intervals for frame notches
    NOTCH_INTERVALS_F = tuple(
        Frame(value)
        for value in [1, 5] + [multiplier * (10**power) for power in range(1, 5) for multiplier in (1, 2, 2.5, 5, 7.5)]
    )

    NOTCH_INTERVALS_T = tuple(
        Time(seconds=n) for n in [1, 2, 5, 10, 15, 30, 60, 90, 120, 300, 600, 900, 1200, 1800, 2700, 3600, 5400, 7200]
    )
    MODES = ("frame", "time")

    BACKGROUND_COLOR = QPalette.ColorRole.Window
    TEXT_COLOR = QPalette.ColorRole.WindowText
    SCROLL_BAR_COLOR = QPalette.ColorRole.WindowText

    HOVER_TIME_FORMAT = "{H:02d}:{M:02d}:{S:02d}.{ms:03d}"
    HOVER_PADDING_H = 6

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        SettingsManager.signals.globalChanged.connect(self._on_settings_changed)

        self._mode: Literal["frame", "time"] = SettingsManager.global_settings.timeline.mode

        self.setAutoFillBackground(True)

        self.rect_f = QRectF()
        self.scroll_rect = QRectF()

        # Visual Metrics (scaled by display_scale)
        self.display_scale = SettingsManager.global_settings.timeline.display_scale
        self.notch_interval_target_x = 75
        self.notch_height = 6
        self.font_height = 10
        self.notch_scroll_interval = 2
        self.scroll_height = 10

        self.set_sizes()

        # Internal cursor state (can be Frame, Time, or raw int pixels)
        self._cursor_val: int | Frame | Time = 0

        self.custom_notches = dict[str, set[CustomNotch[Any]]]()

        # Optimization attributes
        self.setAttribute(Qt.WidgetAttribute.WA_OpaquePaintEvent)
        self.setMouseTracking(True)

        # Interaction state
        self.mousepressed = False
        self.hover_x: int | None = None
        self.is_events_blocked = False

        # Initialize cache
        self.notches_cache = self._init_notches_cache()

        # Hover popup
        self.hover_popup = TimelineHoverPopup(self)

        # Context menu for mode switching
        self.context_menu = QMenu(self)

        # Segmented control for mode selection
        self.mode_selector = SegmentedControl(["Frame", "Time"], self)
        self.mode_selector.index = self.MODES.index(self._mode)
        self.mode_selector.segmentChanged.connect(self._on_mode_segment_changed)

        self.mode_selector_action = QWidgetAction(self.context_menu)
        self.mode_selector_action.setDefaultWidget(self.mode_selector)
        self.context_menu.addAction(self.mode_selector_action)

    @property
    def total_frames(self) -> int:
        if isinstance((parent := self.parent()), TimelineControlBar):
            return parent.total_frames

        raise NotImplementedError

    @property
    def total_time(self) -> Time:
        if isinstance((parent := self.parent()), TimelineControlBar):
            return parent.total_time

        raise NotImplementedError

    @property
    def cum_durations(self) -> list[Time] | None:
        if isinstance((parent := self.parent()), TimelineControlBar):
            return parent.cum_durations

        raise NotImplementedError

    @property
    def cursor_x(self) -> int:
        """Returns the X pixel coordinate of the cursor."""
        return self.cursor_to_x(self._cursor_val)

    @cursor_x.setter
    def cursor_x(self, x: int | Frame | Time) -> None:
        """Sets the cursor value (can be int pixel, Frame, or Time), triggering a redraw."""
        if self.mousepressed:
            return

        self._cursor_val = x
        self.update()

    @property
    def mode(self) -> Literal["frame", "time"]:
        """Current display mode (Frame or Time)."""
        return self._mode

    @mode.setter
    def mode(self, value: Literal["frame", "time"]) -> None:
        """Sets the display mode, triggering a redraw if changed."""
        new_mode = value

        if new_mode == self._mode:
            return

        self._mode = new_mode

        # Update segmented control state
        self.mode_selector.index = self.MODES.index(new_mode)

        self.update()

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        self.context_menu.exec(event.globalPos())

    def paintEvent(self, event: QPaintEvent) -> None:
        self.rect_f = QRectF(self.rect())

        with QPainter(self) as painter:
            self._draw_widget(painter)

    def _draw_widget(self, painter: QPainter) -> None:
        setup_key = Notch.CacheKey(self.rect_f, self.total_frames)

        # Current cache entry for the current mode (Frame or Time)
        cache_entry = self.notches_cache[self.mode]

        # Unpack value components from the cache
        self.scroll_rect, labels_notches, rects_to_draw = cache_entry.value

        # Check if cache needs regeneration (if size or total frames changed)
        if setup_key != cache_entry.key:
            lnotch_y = self.rect_f.top() + self.font_height + self.notch_height + 5
            lnotch_x = self.rect_f.left()
            lnotch_top = lnotch_y - self.notch_height

            labels_notches = list[Notch[Any]]()
            label_format = ""

            # Generate notches based on mode
            if self.mode == "time":
                max_value_t = self.total_time
                notch_interval_t = self.calculate_notch_interval_t(self.notch_interval_target_x)
                label_format = generate_label_format(notch_interval_t, max_value_t)
                label_notch_t = Time()

                # Generate intermediate notches
                if notch_interval_t > Time(seconds=0):
                    while lnotch_x < self.rect_f.right() and label_notch_t <= max_value_t:
                        labels_notches.append(
                            Notch(deepcopy(label_notch_t), line=QLineF(lnotch_x, lnotch_y, lnotch_x, lnotch_top))
                        )
                        label_notch_t = Time(seconds=label_notch_t.total_seconds() + notch_interval_t.total_seconds())
                        lnotch_x = self.cursor_to_x(label_notch_t)

                # Add the final notch at the very end
                end_notch_t = Notch(
                    max_value_t, line=QLineF(self.rect_f.right() - 1, lnotch_y, self.rect_f.right() - 1, lnotch_top)
                )
                labels_notches.append(end_notch_t)

            elif self.mode == "frame":
                max_value_f = Frame(self.total_frames - 1)
                notch_interval_f = self.calculate_notch_interval_f(self.notch_interval_target_x)
                label_notch_f = Frame(0)

                # Generate intermediate notches
                if notch_interval_f > 0:
                    while lnotch_x < self.rect_f.right() and label_notch_f <= max_value_f:
                        labels_notches.append(
                            Notch(deepcopy(label_notch_f), line=QLineF(lnotch_x, lnotch_y, lnotch_x, lnotch_top))
                        )
                        # Ensure arithmetic results in Frame
                        label_notch_f = Frame(label_notch_f + notch_interval_f)
                        lnotch_x = self.cursor_to_x(label_notch_f)

                # Add the final notch at the very end
                end_notch_f = Notch(
                    max_value_f, line=QLineF(self.rect_f.right() - 1, lnotch_y, self.rect_f.right() - 1, lnotch_top)
                )
                labels_notches.append(end_notch_f)
            else:
                raise NotImplementedError

            # Define the scrollable area rectangle
            self.scroll_rect = QRectF(
                self.rect_f.left(), lnotch_y + self.notch_scroll_interval, self.rect_f.width(), self.scroll_height
            )

            # Generate rectangles for text labels to draw
            rects_to_draw = list[tuple[QRectF, str]]()

            for i, notch in enumerate(labels_notches):
                match self.mode:
                    case "frame":
                        label = str(notch.data)
                    case "time" if isinstance(notch.data, Time):
                        label = notch.data.to_ts(label_format)
                    case _:
                        raise NotImplementedError

                anchor_rect = QRectF(notch.line.x2(), notch.line.y2(), 0, 0)

                # Align labels based on their position (first, last, or middle)
                if i == 0:
                    rect = painter.boundingRect(
                        anchor_rect, Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignLeft, label
                    )
                elif i == (len(labels_notches) - 1):
                    rect = painter.boundingRect(
                        anchor_rect, Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignRight, label
                    )
                elif i == (len(labels_notches) - 2):
                    # Special handling for the second to last notch to prevent overlap with the last one
                    rect = painter.boundingRect(
                        anchor_rect, Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignHCenter, label
                    )

                    last_notch = labels_notches[-1]

                    match self.mode:
                        case "frame":
                            last_label = str(last_notch.data)
                        case "time" if isinstance(last_notch.data, Time):
                            last_label = last_notch.data.to_ts(label_format)
                        case _:
                            raise NotImplementedError

                    anchor_rect_last = QRectF(last_notch.line.x2(), last_notch.line.y2(), 0, 0)
                    last_rect = painter.boundingRect(
                        anchor_rect_last, Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignRight, last_label
                    )

                    # If overlap is detected, remove the second to last notch
                    if last_rect.left() - rect.right() < self.notch_interval_target_x / 10:
                        labels_notches.pop(-2)
                        rects_to_draw.append((last_rect, last_label))
                        break
                else:
                    rect = painter.boundingRect(
                        anchor_rect, Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignHCenter, label
                    )

                rects_to_draw.append((rect, label))

            # Update the cache with the new values
            self.notches_cache[self.mode] = Notch.CacheEntry(
                setup_key,
                Notch.CacheValue(self.scroll_rect, labels_notches, rects_to_draw),
            )

        # Define the cursor line position (contained within scroll_rect)
        cursor_line = QLineF(self.cursor_x, self.scroll_rect.top(), self.cursor_x, self.scroll_rect.bottom())

        # DRAWING START

        # Clear background
        painter.fillRect(self.rect_f, self.palette().color(self.BACKGROUND_COLOR))
        painter.setPen(QPen(self.palette().color(self.TEXT_COLOR)))
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Draw text labels
        for txt_rect, text in rects_to_draw:
            painter.drawText(txt_rect, text)

        painter.setRenderHint(QPainter.RenderHint.Antialiasing, False)

        # Draw main notch lines
        for notch in labels_notches:
            painter.drawLine(notch.line)

        # Draw scroll bar area
        painter.fillRect(self.scroll_rect, self.palette().color(self.SCROLL_BAR_COLOR))

        # Draw custom notches from providers (e.g. bookmarks, keyframes)
        for provider_notches in self.custom_notches.values():
            for p_notch in provider_notches:
                p_notch.draw(painter, self.scroll_rect)

        # Draw current frame cursor
        cursor_pen = QPen(self.palette().color(self.BACKGROUND_COLOR), 2)
        cursor_pen.setCosmetic(True)
        painter.setPen(cursor_pen)
        painter.drawLine(cursor_line)

        # Draw hover indicator if mouse is over the widget
        if self.hover_x is not None:
            if self.mode == "frame":
                text = str(self.x_to_frame(self.hover_x))
            else:
                text = self.x_to_time(self.hover_x).to_ts(self.HOVER_TIME_FORMAT)

            painter.setPen(QPen(self.palette().color(self.BACKGROUND_COLOR), 1, Qt.PenStyle.DashLine))
            painter.drawLine(QLineF(self.hover_x, self.scroll_rect.top(), self.hover_x, self.scroll_rect.bottom()))

            fm = painter.fontMetrics()
            text_width = fm.horizontalAdvance(text)
            text_height = fm.height()

            rect_x = self.hover_x - (text_width / 2) - (self.HOVER_PADDING_H / 2)
            if rect_x < 0:
                rect_x = 0
            elif rect_x + text_width + self.HOVER_PADDING_H > self.rect_f.width():
                rect_x = self.rect_f.width() - text_width - self.HOVER_PADDING_H

            rect_y = self.rect_f.top()

            bg_rect = QRectF(rect_x, rect_y, text_width + self.HOVER_PADDING_H, text_height)
            painter.fillRect(bg_rect, self.palette().color(self.BACKGROUND_COLOR))
            painter.setPen(self.palette().color(self.TEXT_COLOR))
            painter.drawText(bg_rect, Qt.AlignmentFlag.AlignCenter, text)

    def moveEvent(self, event: QMoveEvent) -> None:
        super().moveEvent(event)
        self.update()

    def leaveEvent(self, event: QEvent) -> None:
        super().leaveEvent(event)
        self.hover_x = None
        self.hover_popup.hide()
        self.update()

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        if not self.is_events_blocked:
            self.mousepressed = False

    def mousePressEvent(self, event: QMouseEvent) -> None:
        super().mousePressEvent(event)

        if self.is_events_blocked:
            return

        # Only left-click triggers scrubbing; right-click is for context menu
        if event.button() != Qt.MouseButton.LeftButton:
            return

        self.mousepressed = True
        self.mouseMoveEvent(event)

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        super().mouseMoveEvent(event)

        if self.is_events_blocked:
            return

        self.hover_x = int(clamp(event.position().x(), 0, self.rect_f.width()))

        if SettingsManager.global_settings.timeline.view_hover_zoom:
            self.hover_popup.update_state(self.hover_x)
            self.hover_popup.show()
        else:
            self.hover_popup.hide()

        self.update()

        if not self.mousepressed:
            return

        pos = event.position()
        # Check if within scroll area
        scroll_rect = self.notches_cache[self.mode].value.scroll_rect

        # Allow clicking a bit above/below for usability
        click_zone = QRectF(scroll_rect)
        click_zone.setTop(click_zone.top() - 10)
        click_zone.setBottom(click_zone.bottom() + 10)

        if click_zone.contains(pos):
            new_x = int(clamp(pos.x(), 0, self.rect_f.width()))

            self._cursor_val = new_x
            self.update()

            self.clicked.emit(self.x_to_frame(new_x), self.x_to_time(new_x))

    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)
        self.update()

    def set_sizes(self) -> None:
        # Reset cache as sizes have changed
        self.notches_cache = self._init_notches_cache()

        self.notch_interval_target_x = round(75 * self.display_scale)
        self.notch_height = round(6 * self.display_scale)
        self.font_height = round(10 * self.display_scale)
        self.notch_scroll_interval = round(2 * self.display_scale)
        self.scroll_height = round(10 * self.display_scale)

        self.setMinimumWidth(self.notch_interval_target_x)
        self.setFixedHeight(round(33 * self.display_scale))

        font = self.font()
        font.setPixelSize(self.font_height)
        self.setFont(font)

        self.update()

    def calculate_notch_interval_t(self, target_interval_x: int) -> Time:
        margin = 1 + SettingsManager.global_settings.timeline.notches_margin / 100
        target_interval_t = self.x_to_time(target_interval_x)

        for interval in self.NOTCH_INTERVALS_T:
            if target_interval_t < Time(seconds=interval.total_seconds() * margin):
                return interval

        return self.NOTCH_INTERVALS_T[-1]

    def calculate_notch_interval_f(self, target_interval_x: int) -> Frame:
        margin = 1 + SettingsManager.global_settings.timeline.notches_margin / 100
        target_interval_f = self.x_to_frame(target_interval_x)

        for interval in self.NOTCH_INTERVALS_F:
            if target_interval_f < Frame(round(int(interval) * margin)):
                return interval

        return self.NOTCH_INTERVALS_F[-1]

    def x_to_time(self, x: int) -> Time:
        """Converts an X pixel coordinate to a Time value."""
        if self.rect_f.width() == 0:
            return Time()

        if not self.cum_durations:
            return Time()

        return self.cum_durations[frame - 1] if (frame := self.x_to_frame(x)) > 0 else Time()

    def x_to_frame(self, x: int) -> Frame:
        """Converts an X pixel coordinate to a Frame number."""
        return Frame(0) if self.rect_f.width() == 0 else Frame(round(x / self.rect_f.width() * self.total_frames))

    def cursor_to_x(self, cursor: int | Frame | Time) -> int:
        """
        Convert a cursor value (Time, Frame, or int pixel) to an X pixel coordinate.
        """
        try:
            width = self.rect_f.width()

            if isinstance(cursor, Time):
                return (
                    0 if not self.cum_durations else self.cursor_to_x(Frame(bisect_right(self.cum_durations, cursor)))
                )

            if isinstance(cursor, Frame):
                return floor(cursor / self.total_frames * width)

            return cursor

        except (ZeroDivisionError, ValueError):
            return 0

    def add_notch(
        self,
        key: str,
        data: Frame | Time,
        end_data: Frame | Time | None = None,
        color: Qt.GlobalColor | QColor | QRgba64 | str | int = Qt.GlobalColor.black,
        label: str = "",
        id: Hashable | None = None,
    ) -> None:
        cursor_x = self.cursor_to_x(data)
        cursor_line = QLineF(
            cursor_x,
            self.scroll_rect.top(),
            cursor_x,
            self.scroll_rect.top() + self.scroll_rect.height() - 1,
        )

        if end_data is not None:
            end_x = self.cursor_to_x(end_data)
            end_line = QLineF(
                end_x,
                self.scroll_rect.top(),
                end_x,
                self.scroll_rect.top() + self.scroll_rect.height() - 1,
            )
        else:
            end_line = None

        self.custom_notches.setdefault(key, set()).add(
            CustomNotch(
                id or complex_hash.hash(key, data, end_data),
                data,  # pyright: ignore[reportArgumentType]
                end_data,  # pyright: ignore[reportArgumentType]
                color,
                cursor_line,
                end_line,
                label,
            )
        )

    def discard_notch(
        self, key: str, data: Frame | Time, end_data: Frame | Time | None = None, id: Hashable | None = None
    ) -> None:
        self.custom_notches.get(key, set()).discard(
            CustomNotch(
                id or complex_hash.hash(key, data, end_data),
                data,  # pyright: ignore[reportArgumentType]
                end_data,  # pyright: ignore[reportArgumentType]
            )
        )

    @contextmanager
    def block_events(self) -> Iterator[None]:
        self.is_events_blocked = True
        self.mousepressed = False
        try:
            yield
        finally:
            self.is_events_blocked = False

    def _init_notches_cache(self) -> dict[Literal["frame", "time"], Notch.CacheEntry[Any]]:
        return {
            "frame": Notch.CacheEntry(Notch.CacheKey(QRectF(), -1), Notch.CacheValue(QRectF(), [], [])),
            "time": Notch.CacheEntry(Notch.CacheKey(QRectF(), -1), Notch.CacheValue(QRectF(), [], [])),
        }

    def _on_settings_changed(self) -> None:
        self._mode = SettingsManager.global_settings.timeline.mode
        self.display_scale = SettingsManager.global_settings.timeline.display_scale
        self.notches_cache = self._init_notches_cache()
        self.set_sizes()
        self.update()

    def _on_mode_segment_changed(self, index: int) -> None:
        self.mode = "frame" if index == 0 else "time"


class FrameEdit(QSpinBox):
    frameChanged = Signal(Frame, Frame)

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        self.valueChanged.connect(self._on_value_changed)

        self.setMinimum(0)
        self.setKeyboardTracking(False)

        self.old_value = self.value()

    def validate(self, input_text: str, pos: int) -> object:
        if input_text.isdigit():
            val = int(input_text)

            if val > self.maximum():
                max_str = str(self.maximum())
                return QValidator.State.Acceptable, max_str, len(max_str)

        return super().validate(input_text, pos)

    def _on_value_changed(self, value: int) -> None:
        self.frameChanged.emit(Frame(value), Frame(self.old_value))
        self.old_value = value


class TimeEdit(QTimeEdit):
    valueChanged = Signal(QTime, QTime)

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        self.timeChanged.connect(self._on_time_changed)

        self.setDisplayFormat("H:mm:ss.zzz")
        self.setButtonSymbols(QTimeEdit.ButtonSymbols.NoButtons)
        self.setMinimumTime(QTime())
        self.setKeyboardTracking(False)

        self.old_time = self.time()

    def _on_time_changed(self, value: QTime) -> None:
        self.valueChanged.emit(self.time(), self.old_time)
        self.old_time = self.time()


class StepSpinBox(QSpinBox):
    def stepBy(self, steps: int) -> None:
        if self.value() == 1 and steps < 0:
            self.setValue(-1)
        elif self.value() == -1 and steps > 0:
            self.setValue(1)
        else:
            super().stepBy(steps)


@dataclass(slots=True, repr=False, eq=False, match_args=False)
class PlaybackSettings:
    seek_step: int = 1
    speed: float = 1.0
    uncapped: bool = False
    zone_frames: int = 100
    loop: bool = False
    step: int = 1


class PlaybackContainer(QWidget, IconReloadMixin):
    ICON_SIZE = QSize(24, 24)
    ICON_COLOR = QPalette.ColorRole.ToolTipText

    settingsChanged = Signal(int, float, bool)  # seek_step, speed, uncapped
    playZone = Signal(int, bool, int)  # zone_frames, loop, int
    volumeChanged = Signal(float)  # volume 0.0-1.0
    muteChanged = Signal(bool)  # is_muted
    audioDelayChanged = Signal(float)  # seconds

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.setAutoFillBackground(True)

        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Preferred)
        self.setObjectName(self.__class__.__name__)

        self.current_layout = QHBoxLayout(self)
        self.current_layout.setContentsMargins(4, 0, 4, 0)
        self.current_layout.setSpacing(4)

        # Buttons creation
        self.seek_n_back_btn = self._make_button(IconName.REWIND, "Seek N frames backward")
        self.seek_1_back_btn = self._make_button(IconName.SKIP_BACK, "Seek 1 frame backward")

        # Play/Pause uses different icons for each state (not just color)
        self.play_pause_btn = self._make_button(self._make_play_pause_icon(), "Play / Pause", checkable=True)

        self.seek_1_fwd_btn = self._make_button(IconName.SKIP_FORWARD, "Seek 1 frame forward")
        self.seek_n_fwd_btn = self._make_button(IconName.FAST_FORWARD, "Seek N frames forward")

        self.current_layout.addSpacing(5)

        self.time_edit = TimeEdit(self)
        self.current_layout.addWidget(self.time_edit)

        self.frame_edit = FrameEdit(self)
        self.current_layout.addWidget(self.frame_edit)

        self.current_layout.addSpacing(5)

        # Audio controls
        self.audio_controls = QWidget(self)
        self.audio_controls_layout = QHBoxLayout(self.audio_controls)
        self.audio_controls_layout.setContentsMargins(0, 0, 0, 0)
        self.audio_controls_layout.setSpacing(4)

        self.mute_btn = self.make_tool_button(
            self.make_icon((IconName.VOLUME_HIGH, self.palette().color(self.ICON_COLOR)), size=self.ICON_SIZE),
            "Mute / Unmute",
            self.audio_controls,
            checkable=True,
            icon_size=self.ICON_SIZE,
            color_role=self.ICON_COLOR,
        )
        self.mute_btn.clicked.connect(self._on_mute_clicked)
        self.audio_controls_layout.addWidget(self.mute_btn)

        # Volume slider
        self.volume_slider = QSlider(Qt.Orientation.Horizontal, self.audio_controls)
        self.volume_slider.setRange(0, 1000)
        self.volume_slider.setValue(int(SettingsManager.global_settings.playback.default_volume * 1000))
        self.volume_slider.setFixedWidth(60)
        self.volume_slider.setToolTip("Volume: 50%")
        self.volume_slider.valueChanged.connect(self._on_volume_changed)
        self.audio_controls_layout.addWidget(self.volume_slider)

        self.current_layout.addWidget(self.audio_controls)
        self.audio_controls.setEnabled(False)

        self._is_muted = False
        self._volume = SettingsManager.global_settings.playback.default_volume
        self._audio_delay = SettingsManager.global_settings.playback.audio_delay
        self._update_mute_icon()

        self._setup_context_menu()

        # Different icons per state, use custom callback
        self.register_icon_callback(lambda: self.play_pause_btn.setIcon(self._make_play_pause_icon()))
        self.register_icon_callback(self._update_mute_icon)

    def _make_play_pause_icon(self) -> QIcon:
        palette = self.palette()
        return self.make_icon(
            {
                (QIcon.Mode.Normal, QIcon.State.Off): (
                    IconName.PLAY,
                    palette.color(self.ICON_COLOR),
                ),
                (QIcon.Mode.Normal, QIcon.State.On): (
                    IconName.PAUSE,
                    palette.color(QPalette.ColorRole.Mid),
                ),
            },
            size=self.ICON_SIZE,
        )

    def _make_button(
        self,
        icon: IconName | QIcon,
        tooltip: str,
        *,
        checkable: bool = False,
        color: QColor | None = None,
    ) -> QToolButton:
        btn = self.make_tool_button(
            icon,
            tooltip,
            self,
            checkable=checkable,
            icon_size=self.ICON_SIZE,
            color=color,
            color_role=self.ICON_COLOR,
        )
        self.current_layout.addWidget(btn)
        return btn

    def _setup_context_menu(self) -> None:
        self.context_menu = QMenu(self)

        seek_step_widget = QWidget(self.context_menu)
        seek_step_layout = QHBoxLayout(seek_step_widget)
        seek_step_layout.setContentsMargins(8, 4, 8, 4)

        seek_step_layout.addWidget(QLabel("Seek Step", seek_step_widget))

        self.seek_step_spinbox = QSpinBox(seek_step_widget)
        self.seek_step_spinbox.setMinimum(1)
        self.seek_step_spinbox.setMaximum(1_000_000)
        self.seek_step_spinbox.valueChanged.connect(self._on_seek_step_changed)
        seek_step_layout.addWidget(self.seek_step_spinbox)

        seek_step_action = QWidgetAction(self.context_menu)
        seek_step_action.setDefaultWidget(seek_step_widget)
        self.context_menu.addAction(seek_step_action)

        self.reset_seek_step_to_global_action = self.context_menu.addAction("Reset to Global")
        self.reset_seek_step_to_global_action.triggered.connect(self._on_reset_seek_step)

        self.context_menu.addSeparator()

        speed_widget = QWidget(self.context_menu)
        speed_layout = QFormLayout(speed_widget)
        speed_layout.setContentsMargins(6, 6, 6, 6)

        # Speed slider: 0-100 with 1.0x at center (position 50)
        # Left half (0-50): 0.25x to 1.0x in 0.25 steps
        # Right half (51-100): 1.25x to 4.0x in 0.25 steps
        self.speed_slider = QSlider(Qt.Orientation.Horizontal, speed_widget)
        self.speed_slider.setRange(0, 100)
        self.speed_slider.setValue(50)
        self.speed_slider.setMinimumWidth(100)
        self.speed_slider.setToolTip("1.00x")
        self.speed_slider.valueChanged.connect(self._on_speed_slider_changed)
        self.speed_slider_max = 8.0

        self.speed_reset_btn = self.make_tool_button(IconName.ARROW_U_TOP_LEFT, "Reset to 1.0x", speed_widget)
        self.speed_reset_btn.clicked.connect(self._on_reset_speed)

        speed_row = QWidget(speed_widget)
        speed_row_layout = QHBoxLayout(speed_row)
        speed_row_layout.setContentsMargins(0, 0, 0, 0)
        speed_row_layout.setSpacing(4)
        speed_row_layout.addWidget(self.speed_slider)
        speed_row_layout.addWidget(self.speed_reset_btn)

        self.uncap_checkbox = QCheckBox("Uncap FPS")
        self.uncap_checkbox.toggled.connect(self._on_uncap_changed)

        uncap_row = QWidget(speed_widget)
        uncap_row_layout = QHBoxLayout(uncap_row)
        uncap_row_layout.setContentsMargins(0, 0, 0, 0)
        uncap_row_layout.addStretch()
        uncap_row_layout.addWidget(self.uncap_checkbox)

        speed_layout.addRow("Speed Limit", speed_row)
        speed_layout.addRow(uncap_row)

        speed_action = QWidgetAction(self.context_menu)
        speed_action.setDefaultWidget(speed_widget)
        self.context_menu.addAction(speed_action)

        self.context_menu.addSeparator()

        # Zone Playback section
        zone_widget = QWidget(self.context_menu)
        zone_layout = QFormLayout(zone_widget)
        zone_layout.setContentsMargins(6, 6, 6, 6)

        # Time edit
        self.zone_time_edit = TimeEdit(zone_widget)
        self.zone_time_edit.valueChanged.connect(self._on_zone_time_changed)

        # Frame count spinbox
        self.zone_frame_spinbox = FrameEdit(zone_widget)
        self.zone_frame_spinbox.setMinimum(1)
        self.zone_frame_spinbox.setMaximum(1_000_000)
        self.zone_frame_spinbox.setFixedWidth(90)
        self.zone_frame_spinbox.frameChanged.connect(self._on_zone_frames_changed)

        # Row with both edits
        zone_edits_row = QWidget(zone_widget)
        zone_edits_layout = QHBoxLayout(zone_edits_row)
        zone_edits_layout.setContentsMargins(0, 0, 0, 0)
        zone_edits_layout.setSpacing(4)
        zone_edits_layout.addWidget(self.zone_time_edit)
        zone_edits_layout.addWidget(self.zone_frame_spinbox)

        # Play zone button
        self.play_zone_btn = self.make_tool_button(IconName.PLAY, "Play Zone", zone_widget)
        self.play_zone_btn.clicked.connect(self._on_play_zone_clicked)

        # Loop checkbox
        self.loop_checkbox = QCheckBox("Loop")
        self.loop_checkbox.toggled.connect(self._on_loop_changed)

        # Step spinbox
        self.step_frame_spinbox = StepSpinBox(zone_widget, minimum=-1_000_000, maximum=1_000_000, value=1)
        self.step_frame_spinbox.setFixedWidth(90)
        self.step_frame_spinbox.setToolTip("Step size (negative values are allowed)")
        self.step_frame_spinbox.valueChanged.connect(self._on_step_changed)

        # Row with play step
        zone_play_step_row = QWidget(zone_widget)
        zone_play_step_layout = QHBoxLayout(zone_play_step_row)
        zone_play_step_layout.setContentsMargins(0, 0, 0, 0)
        zone_play_step_layout.setSpacing(4)
        zone_play_step_layout.addStretch()
        zone_play_step_layout.addWidget(self.step_frame_spinbox)

        # Row with loop and play button
        zone_controls_row = QWidget(zone_widget)
        zone_controls_layout = QHBoxLayout(zone_controls_row)
        zone_controls_layout.setContentsMargins(0, 0, 0, 0)
        zone_controls_layout.setSpacing(4)
        zone_controls_layout.addStretch()
        zone_controls_layout.addWidget(self.play_zone_btn)
        zone_controls_layout.addWidget(self.loop_checkbox)

        zone_layout.addRow("Zone Time/Frame", zone_edits_row)
        zone_layout.addRow("Play Step", zone_play_step_row)
        zone_layout.addRow(zone_controls_row)

        zone_action = QWidgetAction(self.context_menu)
        zone_action.setDefaultWidget(zone_widget)
        self.context_menu.addAction(zone_action)

        self.settings = PlaybackSettings()

        self.context_menu.addSeparator()

        # Audio
        self.audio_widget = QWidget(self.context_menu)
        audio_layout = QFormLayout(self.audio_widget)
        audio_layout.setContentsMargins(6, 6, 6, 6)

        self.audio_output_combo = QComboBox(self.audio_widget)
        audio_layout.addRow("Audio Output", self.audio_output_combo)

        self.audio_delay_combo = QDoubleSpinBox(
            self.audio_widget,
            suffix=" ms",
            decimals=3,
            minimum=-10000,
            maximum=10000,
            value=self.audio_delay * 1000,
        )
        self.audio_delay_combo.valueChanged.connect(self._on_audio_delay_changed)
        audio_delay_layout = QHBoxLayout()
        audio_delay_layout.addWidget(QLabel("Delay", self.audio_widget))
        audio_delay_layout.addWidget(self.audio_delay_combo)
        audio_layout.addRow(audio_delay_layout)

        audio_action = QWidgetAction(self.context_menu)
        audio_action.setDefaultWidget(self.audio_widget)
        self.context_menu.addAction(audio_action)

        self.reset_audio_delay_to_global_action = self.context_menu.addAction("Reset to Global")
        self.reset_audio_delay_to_global_action.triggered.connect(self._on_reset_audio_delay)

    def _update_mute_icon(self) -> None:
        if self._is_muted:
            self.mute_btn.setIcon(
                self.make_icon(
                    (IconName.VOLUME_MUTE, self.palette().color(QPalette.ColorRole.Mid)),
                    size=self.ICON_SIZE,
                )
            )
            return

        if self._volume == 0:
            icon_name = IconName.VOLUME_OFF
        elif self._volume < 0.33:
            icon_name = IconName.VOLUME_LOW
        elif self._volume < 0.67:
            icon_name = IconName.VOLUME_MID
        else:
            icon_name = IconName.VOLUME_HIGH

        self.mute_btn.setIcon(self.make_icon((icon_name, self.palette().color(self.ICON_COLOR)), size=self.ICON_SIZE))

    @property
    def volume(self) -> float:
        return 0.0 if self._is_muted else self._volume

    @volume.setter
    def volume(self, value: float) -> None:
        self._volume = clamp(value, 0.0, 1.0)

        with QSignalBlocker(self.volume_slider):
            self.volume_slider.setValue(round(self._volume * 1000))

        self._update_mute_icon()

    @property
    def raw_volume(self) -> float:
        return self._volume

    @property
    def is_muted(self) -> bool:
        return self._is_muted

    @is_muted.setter
    def is_muted(self, value: bool) -> None:
        self._is_muted = value

        with QSignalBlocker(self.mute_btn):
            self.mute_btn.setChecked(value)

        self._update_mute_icon()

    @property
    def audio_delay(self) -> float:
        return self._audio_delay

    @audio_delay.setter
    def audio_delay(self, value: float) -> None:
        if self._audio_delay == value:
            return

        self._audio_delay = value

        with QSignalBlocker(self.audio_delay_combo):
            self.audio_delay_combo.setValue(value * 1000)

        self.audioDelayChanged.emit(value)

    @property
    def cum_durations(self) -> list[Time] | None:
        if isinstance((parent := self.parent()), TimelineControlBar):
            return parent.cum_durations

        raise NotImplementedError

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        with QSignalBlocker(self.seek_step_spinbox):
            self.seek_step_spinbox.setValue(self.settings.seek_step)

        with QSignalBlocker(self.speed_slider):
            self.speed_slider.setValue(self._speed_to_slider(self.settings.speed))
            self.speed_slider.setToolTip(f"{self.settings.speed:.2f}x")

        with QSignalBlocker(self.uncap_checkbox):
            self.uncap_checkbox.setChecked(self.settings.uncapped)

        self.speed_slider.setEnabled(not self.settings.uncapped)
        self.speed_reset_btn.setEnabled(not self.settings.uncapped)

        with QSignalBlocker(self.zone_frame_spinbox):
            self.zone_frame_spinbox.setValue(self.settings.zone_frames)

        if self.cum_durations:
            with QSignalBlocker(self.zone_time_edit):
                self.zone_time_edit.setTime(
                    self.cum_durations[self.settings.zone_frames - 1].to_qtime()
                    if self.settings.zone_frames > 0
                    else QTime()
                )

        with QSignalBlocker(self.step_frame_spinbox):
            self.step_frame_spinbox.setValue(self.settings.step)

        with QSignalBlocker(self.loop_checkbox):
            self.loop_checkbox.setChecked(self.settings.loop)

        self.reset_seek_step_to_global_action.setEnabled(
            self.settings.seek_step != SettingsManager.global_settings.timeline.seek_step
        )
        self.reset_audio_delay_to_global_action.setEnabled(
            self.audio_delay != SettingsManager.global_settings.playback.audio_delay
        )

        self.audio_widget.setEnabled(self.audio_output_combo.count() > 0)

        menu_pos = event.globalPos()
        menu_pos.setY(menu_pos.y() - self.context_menu.sizeHint().height())
        self.context_menu.exec(menu_pos)

    @run_in_loop
    def set_audio_outputs(self, aoutputs: list[AudioOutput], index: int | None = None) -> None:
        with QSignalBlocker(self.audio_output_combo):
            self.audio_output_combo.clear()
            self.audio_output_combo.addItems(
                [f"{a.vs_index}: {a.vs_name} ({a.chanels_layout.pretty_name})" for a in aoutputs]
            )

        if len(aoutputs) > 0:
            self.audio_controls.setEnabled(True)

            if index is not None:
                self.audio_output_combo.setCurrentIndex(index)

            self.audioDelayChanged.emit(self.audio_delay)

    def _on_seek_step_changed(self, value: int) -> None:
        self.settings.seek_step = value
        self.reset_seek_step_to_global_action.setEnabled(value != SettingsManager.global_settings.timeline.seek_step)
        self._emit_settings()

    def _on_reset_seek_step(self) -> None:
        # Update spinbox to show global value
        global_seek = SettingsManager.global_settings.timeline.seek_step
        self.settings.seek_step = global_seek

        with QSignalBlocker(self.seek_step_spinbox):
            self.seek_step_spinbox.setValue(global_seek)

        self._emit_settings()

    def _slider_to_speed(self, slider_val: int) -> float:
        # Slider 0-50 -> 0.25 to 1.0 (steps: 0.25, 0.50, 0.75, 1.00)
        # Slider 51-100 -> 1.25 to max (steps: 1.25, 1.50, ..., max)
        speed = (
            0.25 + slider_val / 50.0 * 0.75
            if slider_val <= 50
            else 1.0 + (slider_val - 50) / 50.0 * (self.speed_slider_max - 1.0)
        )
        return round(speed * 4) / 4

    def _speed_to_slider(self, speed: float) -> int:
        return (
            round(((speed - 0.25) / 0.75) * 50)
            if speed <= 1.0
            else round(50 + ((speed - 1.0) / (self.speed_slider_max - 1.0)) * 50)
        )

    def _on_speed_slider_changed(self, value: int) -> None:
        self.settings.speed = self._slider_to_speed(value)
        speed_text = f"{self.settings.speed:.2f}x"
        self.speed_slider.setToolTip(speed_text)

        QToolTip.showText(QCursor.pos(), speed_text, self.speed_slider)

        self._emit_settings()

    def _on_reset_speed(self) -> None:
        self.settings.speed = 1.0

        with QSignalBlocker(self.speed_slider):
            self.speed_slider.setValue(50)

        self.speed_slider.setToolTip("1.00x")
        self._emit_settings()

    def _on_uncap_changed(self, checked: bool) -> None:
        self.settings.uncapped = checked
        self.speed_slider.setEnabled(not checked)
        self.speed_reset_btn.setEnabled(not checked)
        self._emit_settings()

    def _emit_settings(self) -> None:
        self.settingsChanged.emit(self.settings.seek_step, self.settings.speed, self.settings.uncapped)

    def _on_zone_frames_changed(self, new_frame: int, old_frame: int) -> None:
        self.settings.zone_frames = new_frame

        # Convert frames to time
        if self.cum_durations:
            with QSignalBlocker(self.zone_time_edit):
                self.zone_time_edit.setTime(self.cum_durations[new_frame - 1].to_qtime() if new_frame > 0 else QTime())

    def _on_zone_time_changed(self, new_time: QTime, old_time: QTime) -> None:
        # Convert time to frames
        if self.cum_durations:
            frames = max(1, bisect_right(self.cum_durations, Time.from_qtime(new_time)))

            self.settings.zone_frames = frames

            with QSignalBlocker(self.zone_frame_spinbox):
                self.zone_frame_spinbox.setValue(frames)

    def _on_play_zone_clicked(self) -> None:
        self.playZone.emit(self.settings.zone_frames, self.settings.loop, self.settings.step)
        self.context_menu.close()

    def _on_loop_changed(self, checked: bool) -> None:
        self.settings.loop = checked

    def _on_step_changed(self, value: int) -> None:
        if value == 0:
            new_value = 1 if self.settings.step < 0 else -1

            with QSignalBlocker(self.step_frame_spinbox):
                self.step_frame_spinbox.setValue(new_value)

            self.settings.step = new_value
            return

        self.settings.step = value

    def _on_mute_clicked(self, checked: bool) -> None:
        self._is_muted = checked
        self._update_mute_icon()
        self.muteChanged.emit(self._is_muted)

    def _on_volume_changed(self, value: int) -> None:
        self._volume = value / 1000.0
        volume_text = f"Volume: {self._volume * 100:.0f}%"
        self.volume_slider.setToolTip(volume_text)

        QToolTip.showText(QCursor.pos(), volume_text, self.volume_slider)

        self._update_mute_icon()

        # Unmute if volume is changed while muted
        if self._is_muted and self._volume > 0:
            self._is_muted = False
            self.mute_btn.setChecked(False)
            self.muteChanged.emit(False)

        self.volumeChanged.emit(self._volume)

    def _on_audio_delay_changed(self, value: float) -> None:
        self._audio_delay = value / 1000
        self.reset_audio_delay_to_global_action.setEnabled(
            self._audio_delay != SettingsManager.global_settings.playback.audio_delay
        )
        self.audioDelayChanged.emit(self._audio_delay)

    def _on_reset_audio_delay(self) -> None:
        global_delay = SettingsManager.global_settings.playback.audio_delay
        self._audio_delay = global_delay

        with QSignalBlocker(self.audio_delay_combo):
            self.audio_delay_combo.setValue(global_delay * 1000)

        self.audioDelayChanged.emit(global_delay)


class TimelineControlBar(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        # Prevent vertical expansion when window is maximized
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Maximum)

        # Timeline and Playback Controls
        self.timeline_layout = QHBoxLayout(self)
        self.timeline_layout.setContentsMargins(0, 0, 0, 0)
        self.timeline_layout.setSpacing(0)

        # Playback Controls
        self.playback_container = PlaybackContainer(self)

        self.timeline_layout.addWidget(self.playback_container)

        self.timeline = Timeline(self)
        self.timeline_layout.addWidget(self.timeline)

    @property
    def total_frames(self) -> int:
        return self._total_frames

    @property
    def total_time(self) -> Time:
        return self._cum_durations[-1] if self._cum_durations else Time()

    @property
    def cum_durations(self) -> list[Time] | None:
        return self._cum_durations

    def set_data(self, total_frames: int, cum_durations: list[float] | None = None) -> None:
        self._total_frames = total_frames
        self._cum_durations = [Time(seconds=cum) for cum in cum_durations] if cum_durations else None

        self.timeline.update()

        # Playback Container
        with QSignalBlocker(self.playback_container.zone_frame_spinbox):
            self.playback_container.zone_frame_spinbox.setMaximum(Frame(total_frames - 1))

        with QSignalBlocker(self.playback_container.zone_time_edit):
            self.playback_container.zone_time_edit.setMaximumTime(self.total_time.to_qtime())

        with QSignalBlocker(self.playback_container.frame_edit):
            self.playback_container.frame_edit.setMaximum(Frame(total_frames - 1))

        with QSignalBlocker(self.playback_container.time_edit):
            self.playback_container.time_edit.setMaximumTime(self.total_time.to_qtime())

    @run_in_loop(return_future=False)
    def set_playback_controls_enabled(self, enabled: bool) -> None:
        """
        Enable or disable playback controls (except play/pause button).

        During playback, seek buttons and time/frame edits should be disabled,
        but the play/pause button must remain clickable so users can stop playback.
        """
        self.playback_container.seek_n_back_btn.setEnabled(enabled)
        self.playback_container.seek_1_back_btn.setEnabled(enabled)
        self.playback_container.seek_1_fwd_btn.setEnabled(enabled)
        self.playback_container.seek_n_fwd_btn.setEnabled(enabled)
        self.playback_container.time_edit.setEnabled(enabled)
        self.playback_container.frame_edit.setEnabled(enabled)

    @contextmanager
    def disabled(self) -> Iterator[None]:
        """
        Context manager to disable the control bar and block timeline events.

        Disables the widget, blocks mouse events on the timeline, and re-enables on exit.
        If an exception occurs, the toolbar stays disabled.

        Use for single-frame renders, not for continuous playback (use set_playback_controls_enabled instead).
        """
        loop = get_loop()

        @run_in_loop(return_future=False)
        def disable() -> None:
            self.timeline.hover_popup.hide()
            self.setEnabled(False)

        disable()

        # Keep toolbar disabled on error
        with self.timeline.block_events():
            yield

        loop.from_thread(self.setEnabled, True).result()



"""
Icon provider abstraction for vsview.

Provides a unified interface for loading icons from different providers with normalized icon names.
"""

from abc import ABC, abstractmethod
from enum import StrEnum
from importlib import resources
from importlib.resources.abc import Traversable
from typing import ClassVar


class IconName(StrEnum):
    PLAY = "play"
    PAUSE = "pause"
    FAST_FORWARD = "fast-forward"
    REWIND = "rewind"
    SKIP_FORWARD = "skip-forward"
    SKIP_BACK = "skip-back"
    CHECK = "check"
    SPINNER_GAP = "spinner-gap"
    UNLINK = "link-break"
    LINK = "link"
    LINK_2 = "link-simple-horizontal"
    LINK_3 = "link-simple-horizontal-break"
    MAGNIFYING_GLASS = "magnifying-glass"
    FRAME_CORNERS = "frame-corners"
    ARROWS_OUT_CARDINAL = "arrows-out-cardinal"
    ARROW_U_TOP_LEFT = "arrow-u-up-left"
    X_CIRCLE = "x-circle"
    FILE_TEXT = "file-text"
    FILE_VIDEO = "file-video"
    FILE_CODE = "file-code"
    SAVE = "floppy-disk"
    ARROW_LEFT = "arrow-left"
    ARROW_RIGHT = "arrow-right"
    # Audio icons
    VOLUME_HIGH = "speaker-high"
    VOLUME_MID = "speaker-low"
    VOLUME_LOW = "speaker-none"
    VOLUME_OFF = "speaker-x"
    VOLUME_MUTE = "speaker-slash"
    #
    PIPETTE = "eyedropper"
    CLIPBOARD = "clipboard"
    PLUS = "plus"
    MINUS = "minus"
    FILE_IMPORT = "file-arrow-down"
    FILE_EXPORT = "file-arrow-up"
    # Scening icons
    SCENE_ADD = "selection-plus"
    MARK_IN = "selection-background"
    MARK_OUT = "selection-foreground"
    FRAME_ADD = "selection"
    SIDEBAR = "sidebar-simple"


class IconProvider(ABC):
    """Abstract base class for icon providers."""

    @property
    @abstractmethod
    def id(self) -> str:
        """Provider identifier used in settings."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Human-readable provider name for UI."""

    @property
    @abstractmethod
    def weights(self) -> frozenset[str]:
        """Available weight variants for this provider."""

    @property
    @abstractmethod
    def default_weight(self) -> str:
        """Default weight for this provider."""

    @abstractmethod
    def get_folder(self) -> Traversable:
        """Get the assets folder for this provider."""

    @abstractmethod
    def map_name(self, name: IconName) -> str:
        """Map normalized IconName to provider-specific filename (without extension)."""

    def get_icon_path(self, name: IconName, weight: str) -> Traversable:
        """Get the path to a specific icon file, with fallback to default weight."""
        mapped = self.map_name(name)
        folder = self.get_folder()

        # Try requested weight first
        if weight != self.default_weight:
            filename = f"{mapped}-{weight}.svg"
            path = folder / filename
            # Fallback to default weight if variant doesn't exist
            if not path.is_file():
                path = folder / f"{mapped}.svg"
            return path

        return folder / f"{mapped}.svg"


class PhosphorProvider(IconProvider):
    """Phosphor Icons provider."""

    @property
    def id(self) -> str:
        return "phosphor"

    @property
    def name(self) -> str:
        return "Phosphor Icons"

    @property
    def weights(self) -> frozenset[str]:
        return frozenset({"thin", "light", "regular", "bold", "fill", "duotone"})

    @property
    def default_weight(self) -> str:
        return "regular"

    def get_folder(self) -> Traversable:
        return resources.files("vsview.assets.icons.phosphor")

    def map_name(self, name: IconName) -> str:
        # Phosphor uses same kebab-case names as IconName
        return name.value


class MaterialProvider(IconProvider):
    """Material Design Icons provider."""

    # Mapping from normalized IconName to Material icon names
    NAME_MAP: ClassVar[dict[IconName, str]] = {
        IconName.SKIP_FORWARD: "skip-next",
        IconName.SKIP_BACK: "skip-previous",
        IconName.SPINNER_GAP: "loading",
        IconName.UNLINK: "link-variant-off",
        IconName.LINK: "link-variant",
        IconName.LINK_2: "link-variant-plus",
        IconName.LINK_3: "link-variant-minus",
        IconName.MAGNIFYING_GLASS: "magnify",
        IconName.FRAME_CORNERS: "border-radius",
        IconName.ARROWS_OUT_CARDINAL: "arrow-expand-all",
        IconName.ARROW_U_TOP_LEFT: "arrow-u-left-top",
        IconName.X_CIRCLE: "close-circle",
        IconName.FILE_TEXT: "file-document",
        IconName.SAVE: "content-save",
        IconName.VOLUME_HIGH: "volume-high",
        IconName.VOLUME_MID: "volume-medium",
        IconName.VOLUME_LOW: "volume-low",
        IconName.VOLUME_OFF: "volume-mute",
        IconName.VOLUME_MUTE: "volume-off",
        IconName.FILE_IMPORT: "file-import",
        IconName.FILE_EXPORT: "file-export",
        IconName.SCENE_ADD: "plus-box-multiple",
        IconName.MARK_IN: "step-backward",
        IconName.MARK_OUT: "step-forward",
        IconName.FRAME_ADD: "plus-box",
        IconName.SIDEBAR: "page-layout-sidebar-right",
    }

    @property
    def id(self) -> str:
        return "material"

    @property
    def name(self) -> str:
        return "Material Design Icons"

    @property
    def weights(self) -> frozenset[str]:
        return frozenset({"regular", "outline"})

    @property
    def default_weight(self) -> str:
        return "regular"

    def get_folder(self) -> Traversable:
        return resources.files("vsview.assets.icons.material")

    def map_name(self, name: IconName) -> str:
        return self.NAME_MAP.get(name, name.value)


class LucideProvider(IconProvider):
    """Lucide Icons provider."""

    # Mapping from normalized IconName to Lucide icon names
    NAME_MAP: ClassVar[dict[IconName, str]] = {
        IconName.SPINNER_GAP: "loader-circle",
        IconName.UNLINK: "unlink",
        IconName.LINK_2: "link-2",
        IconName.LINK_3: "unlink-2",
        IconName.MAGNIFYING_GLASS: "search-code",
        IconName.FRAME_CORNERS: "scan",
        IconName.ARROWS_OUT_CARDINAL: "move",
        IconName.ARROW_U_TOP_LEFT: "undo-2",
        IconName.X_CIRCLE: "circle-x",
        IconName.FILE_VIDEO: "file-video-camera",
        IconName.SAVE: "save",
        IconName.VOLUME_HIGH: "volume-2",
        IconName.VOLUME_MID: "volume-1",
        IconName.VOLUME_LOW: "volume",
        IconName.VOLUME_OFF: "volume-x",
        IconName.VOLUME_MUTE: "volume-off",
        IconName.PIPETTE: "pipette",
        IconName.FILE_IMPORT: "file-down",
        IconName.FILE_EXPORT: "file-up",
        IconName.SCENE_ADD: "square-plus",
        IconName.MARK_IN: "square-chevron-left",
        IconName.MARK_OUT: "square-chevron-right",
        IconName.FRAME_ADD: "square-arrow-right-enter",
        IconName.SIDEBAR: "panel-right-open",
    }

    @property
    def id(self) -> str:
        return "lucide"

    @property
    def name(self) -> str:
        return "Lucide Icons"

    @property
    def weights(self) -> frozenset[str]:
        return frozenset({"regular"})

    @property
    def default_weight(self) -> str:
        return "regular"

    def get_folder(self) -> Traversable:
        return resources.files("vsview.assets.icons.lucide")

    def map_name(self, name: IconName) -> str:
        return self.NAME_MAP.get(name, name.value)


# Provider registry
ICON_PROVIDERS: dict[str, IconProvider] = {
    "phosphor": PhosphorProvider(),
    "material": MaterialProvider(),
    "lucide": LucideProvider(),
}



from typing import Any

from vsview.api import WidgetPluginBase
from vsview.api import hookimpl as _vsview_hookimpl

from .plugin import FramePropsPlugin


@_vsview_hookimpl(tryfirst=True)
def vsview_register_toolpanel() -> type[WidgetPluginBase[Any, Any]]:
    return FramePropsPlugin


@_vsview_hookimpl(tryfirst=True)
def vsview_register_tooldock() -> type[WidgetPluginBase[Any, Any]]:
    return FramePropsPlugin



from .base import BaseWorkspace
from .file import GenericFileWorkspace, PythonScriptWorkspace, VideoFileWorkspace
from .loader import LoaderWorkspace, VSEngineWorkspace
from .quick_script import QuickScriptWorkspace

__all__ = [
    "BaseWorkspace",
    "GenericFileWorkspace",
    "LoaderWorkspace",
    "PythonScriptWorkspace",
    "QuickScriptWorkspace",
    "VSEngineWorkspace",
    "VideoFileWorkspace",
]



from typing import Any, cast

import pytest
from pydantic import BaseModel
from pytest_mock import MockerFixture

from vsview.app.plugins._interface import _PluginSettingsStore, _SettingsProxy
from vsview.app.plugins.api import LocalSettingsModel


# Mock models for testing
class MockNestedModel(BaseModel):
    nest_attr: str = "nest_default"


class MockParentModel(BaseModel):
    attr1: str = "default1"
    nest: MockNestedModel = MockNestedModel()


class MockLevel3(BaseModel):
    val: int = 3


class MockLevel2(BaseModel):
    l3: MockLevel3 = MockLevel3()


class MockLevel1(BaseModel):
    l2: MockLevel2 = MockLevel2()


class MockGlobalSettings(BaseModel):
    attr1: str = "default1"
    attr2: int = 10


class MockLocalSettings(LocalSettingsModel):
    attr1: str | None = None
    attr2: int | None = None

    def resolve(self, global_settings: BaseModel) -> "MockLocalSettings":
        resolved = self.model_copy()
        if resolved.attr1 is None:
            resolved.attr1 = getattr(global_settings, "attr1", "default1")
        if resolved.attr2 is None:
            resolved.attr2 = getattr(global_settings, "attr2", 10)
        return resolved


# --- _SettingsProxy Tests ---


def test_proxy_delegation(mocker: MockerFixture) -> None:
    model = MockGlobalSettings()
    on_update = mocker.stub()
    proxy = _SettingsProxy(model, on_update)

    assert proxy.attr1 == "default1"
    assert proxy.attr2 == 10
    on_update.assert_not_called()


def test_proxy_setattr(mocker: MockerFixture) -> None:
    model = MockGlobalSettings()
    on_update = mocker.stub()
    proxy = _SettingsProxy(model, on_update)

    proxy.attr1 = "new_value"
    on_update.assert_called_once_with("attr1", "new_value")


def test_proxy_repr(mocker: MockerFixture) -> None:
    model = MockGlobalSettings()
    proxy = _SettingsProxy(model, mocker.stub())
    assert repr(proxy) == repr(model)


def test_proxy_equality(mocker: MockerFixture) -> None:
    model1 = MockGlobalSettings(attr1="a")
    model2 = MockGlobalSettings(attr1="a")
    model3 = MockGlobalSettings(attr1="b")

    proxy1 = _SettingsProxy(model1, mocker.stub())
    proxy2 = _SettingsProxy(model2, mocker.stub())

    assert proxy1 == proxy2
    assert proxy1 == model1
    assert proxy1 != model3
    assert proxy1 != "not a proxy"


def test_proxy_slots(mocker: MockerFixture) -> None:
    model = MockGlobalSettings()
    proxy = _SettingsProxy(model, mocker.stub())

    # Verify __slots__ exist
    assert hasattr(proxy, "__slots__")
    assert "_model" in proxy.__slots__
    assert "_on_update" in proxy.__slots__

    # Attempting to set an arbitrary attribute should fail
    with pytest.raises(AttributeError):
        proxy.unknown_attr = 5


# --- _PluginSettingsStore Tests ---


@pytest.fixture
def mock_workspace(mocker: MockerFixture) -> Any:
    return mocker.MagicMock()


@pytest.fixture
def mock_plugin(mocker: MockerFixture) -> Any:
    plugin = mocker.MagicMock()
    plugin.identifier = "test_plugin"
    plugin.global_settings_model = MockGlobalSettings
    plugin.local_settings_model = MockLocalSettings
    return plugin


def test_store_get_caching(mocker: MockerFixture, mock_workspace: Any, mock_plugin: Any) -> None:
    mock_settings_manager = mocker.patch("vsview.app.plugins._interface.SettingsManager")
    store = _PluginSettingsStore(mock_workspace)

    # Mock raw settings
    mock_settings_manager.global_settings.plugins = {"test_plugin": {"attr1": "from_storage"}}

    # First call should fetch and validate
    settings1 = store.get(mock_plugin, "global")
    assert settings1 is not None
    assert getattr(settings1, "attr1") == "from_storage"

    # Modify storage - cache should still return old value
    mock_settings_manager.global_settings.plugins["test_plugin"] = {"attr1": "changed"}
    settings2 = store.get(mock_plugin, "global")
    assert settings2 is settings1
    assert getattr(settings2, "attr1") == "from_storage"


def test_store_local_global_resolution(mocker: MockerFixture, mock_workspace: Any, mock_plugin: Any) -> None:
    mock_settings_manager = mocker.patch("vsview.app.plugins._interface.SettingsManager")
    store = _PluginSettingsStore(mock_workspace)

    # Setup global and local storage
    mock_settings_manager.global_settings.plugins = {"test_plugin": {"attr1": "global_val", "attr2": 100}}
    mock_settings_manager.get_local_settings.return_value.plugins = {"test_plugin": {"attr2": 200}}

    # Mock file_path for local settings
    mocker.patch.object(_PluginSettingsStore, "file_path", "test.vpy")

    local_settings = store.get(mock_plugin, "local")
    assert local_settings is not None

    # attr1 should fall back to global, attr2 should be local
    assert getattr(local_settings, "attr1") == "global_val"
    assert getattr(local_settings, "attr2") == 200


def test_store_update_invalidates_cache(mocker: MockerFixture, mock_workspace: Any, mock_plugin: Any) -> None:
    mock_settings_manager = mocker.patch("vsview.app.plugins._interface.SettingsManager")
    store = _PluginSettingsStore(mock_workspace)

    mock_settings_manager.global_settings.plugins = {"test_plugin": {"attr1": "old"}}

    # Populate cache
    settings_old = store.get(mock_plugin, "global")
    assert settings_old is not None
    assert getattr(settings_old, "attr1") == "old"

    # Update through store
    store.update(mock_plugin, "global", attr1="new")

    # Verify persistence
    persisted = mock_settings_manager.global_settings.plugins["test_plugin"]
    if hasattr(persisted, "attr1"):
        assert persisted.attr1 == "new"  # pyright: ignore[reportAttributeAccessIssue]
    else:
        assert persisted["attr1"] == "new"

    # Verify cache invalidation - next get should be fresh
    settings_new = store.get(mock_plugin, "global")
    assert settings_new is not settings_old
    assert settings_new is not None
    assert getattr(settings_new, "attr1") == "new"


def test_store_invalidate(mocker: MockerFixture, mock_workspace: Any, mock_plugin: Any) -> None:
    mock_settings_manager = mocker.patch("vsview.app.plugins._interface.SettingsManager")
    mock_settings_manager.global_settings.plugins = {"test_plugin": {"attr1": "val"}}
    store = _PluginSettingsStore(mock_workspace)

    store.get(mock_plugin, "global")
    # Accessing private cache for verification
    assert mock_plugin in store._caches["global"]

    store.invalidate("global")
    assert mock_plugin not in store._caches["global"]


# --- Integration Test ---


def test_plugin_settings_reactive_write(mocker: MockerFixture, mock_workspace: Any, mock_plugin: Any) -> None:
    mock_settings_manager = mocker.patch("vsview.app.plugins._interface.SettingsManager")
    store = _PluginSettingsStore(mock_workspace)

    def get_proxy() -> _SettingsProxy[MockGlobalSettings]:
        model = cast(MockGlobalSettings, store.get(mock_plugin, "global"))
        return _SettingsProxy(model, lambda k, v: store.update(mock_plugin, "global", **{k: v}))

    mock_settings_manager.global_settings.plugins = {"test_plugin": {"attr1": "initial"}}

    proxy = get_proxy()
    assert proxy.attr1 == "initial"

    # Reactive write
    proxy.attr1 = "updated"

    # Verify persistence call
    assert mock_settings_manager.global_settings.plugins["test_plugin"].attr1 == "updated"  # pyright: ignore[reportAttributeAccessIssue]

    # Next read should be fresh due to cache invalidation in update()
    new_proxy = get_proxy()
    assert new_proxy.attr1 == "updated"
    assert new_proxy is not proxy  # Different model instance


def test_proxy_nested_model_update(mocker: MockerFixture) -> None:
    model = MockParentModel()
    on_update = mocker.stub()
    proxy = _SettingsProxy(model, on_update)

    proxy.nest.nest_attr = "new_nest_value"

    assert model.nest.nest_attr == "new_nest_value"
    on_update.assert_called_once()
    on_update.assert_called_once_with("nest", model.nest)


def test_proxy_deeply_nested_model_update(mocker: MockerFixture) -> None:
    model = MockLevel1()
    on_update = mocker.stub()
    proxy = _SettingsProxy(model, on_update)

    proxy.l2.l3.val = 42

    assert model.l2.l3.val == 42
    on_update.assert_called_once_with("l2", model.l2)



"""Ported back from vstools functions"""

import vapoursynth as vs


def get_chroma_offsets(frame: vs.VideoFrame) -> tuple[float, float]:
    off_top, off_left = 0.0, 0.0
    chromaloc = frame.props.get("_ChromaLocation", 0)

    if chromaloc in [vs.CHROMA_TOP_LEFT, vs.CHROMA_TOP]:
        off_top = 2 ** (frame.format.subsampling_h - 1) - 0.5

    if chromaloc in [vs.CHROMA_LEFT, vs.CHROMA_TOP_LEFT, vs.CHROMA_BOTTOM_LEFT]:
        off_left = 2 ** (frame.format.subsampling_w - 1) - 0.5

    return off_top, off_left


def get_lowest_value(fmt: vs.VideoFormat, chroma: bool, range_in: vs.ColorRange) -> float:
    if fmt.color_family == vs.RGB:
        chroma = False

    if fmt.sample_type is vs.FLOAT:
        return -0.5 if chroma else 0.0

    if range_in == vs.ColorRange.RANGE_LIMITED:
        return 16 << fmt.bits_per_sample - 8

    return 0


def get_peak_value(fmt: vs.VideoFormat, chroma: bool, range_in: vs.ColorRange) -> float:
    if fmt.color_family == vs.RGB:
        chroma = False

    if fmt.sample_type is vs.FLOAT:
        return 0.5 if chroma else 1.0

    if range_in == vs.ColorRange.RANGE_LIMITED:
        return (240 if chroma else 235) << fmt.bits_per_sample - 8

    return (1 << fmt.bits_per_sample) - 1


def scale_value_to_float(value: float, input_frame: vs.VideoFrame, chroma: bool = False) -> float:
    out_value = float(value)

    in_fmt = input_frame.format
    out_fmt = in_fmt.replace(sample_type=vs.FLOAT, bits_per_sample=32)

    prop = input_frame.props.get("_ColorRange")

    if prop is not None:
        range_in = vs.ColorRange(prop)  # type: ignore[arg-type]
    elif in_fmt.color_family == vs.RGB:
        range_in = vs.ColorRange.RANGE_FULL
    else:
        range_in = vs.ColorRange.RANGE_LIMITED

    if input_frame.format.bits_per_sample == 32 or (
        input_frame.format.bits_per_sample == 16 and input_frame.format.sample_type == vs.FLOAT
    ):
        return out_value

    if in_fmt.color_family == vs.RGB:
        chroma = False

    input_peak = get_peak_value(in_fmt, chroma, range_in)
    input_lowest = get_lowest_value(in_fmt, chroma, range_in)
    output_peak = get_peak_value(out_fmt, chroma, range_in)
    output_lowest = get_lowest_value(out_fmt, chroma, range_in)

    if in_fmt.sample_type is vs.INTEGER:
        if chroma:
            out_value -= 128 << (in_fmt.bits_per_sample - 8)
        elif range_in == vs.ColorRange.RANGE_LIMITED:
            out_value -= 16 << (in_fmt.bits_per_sample - 8)

    out_value *= (output_peak - output_lowest) / (input_peak - input_lowest)

    return out_value



from __future__ import annotations

from enum import Enum

from vapoursynth import (
    ChromaLocation,
    ColorPrimaries,
    ColorRange,
    FieldBased,
    MatrixCoefficients,
    TransferCharacteristics,
)

from ..categories import CategoryMatcher
from ..formatters import FormatterProperty

VIDEO_CATEGORY = CategoryMatcher(
    name="Video",
    priority=150,
    order=10,
    exact_matches={
        # Colorimetry
        "_ChromaLocation",
        "_ColorRange",
        "_Matrix",
        "_Transfer",
        "_Primaries",
        # Others
        "_Alpha",
        "_PictType",
        "_DurationNum",
        "_DurationDen",
        "_AbsoluteTime",
        "_SARNum",
        "_SARDen",
        "_FieldBased",
    },
)


def _format_enum(value: int, enum: type[Enum]) -> str:
    return enum(value).name.split("_", 1)[1:][0]


VIDEO_FORMATTERS: list[FormatterProperty] = [
    # Enum-based formatters
    FormatterProperty(
        prop_key="_ChromaLocation",
        value_formatter=lambda v: _format_enum(v, ChromaLocation).title(),
    ),
    FormatterProperty(
        prop_key="_ColorRange",
        value_formatter=lambda v: _format_enum(v, ColorRange).title(),
    ),
    FormatterProperty(
        prop_key="_Matrix",
        value_formatter=lambda v: _format_enum(v, MatrixCoefficients),
    ),
    FormatterProperty(
        prop_key="_Transfer",
        value_formatter=lambda v: _format_enum(v, TransferCharacteristics),
    ),
    FormatterProperty(
        prop_key="_Primaries",
        value_formatter=lambda v: _format_enum(v, ColorPrimaries),
    ),
    FormatterProperty(
        prop_key="_FieldBased",
        value_formatter=lambda v: _format_enum(v, FieldBased).title(),
    ),
    # Others
    FormatterProperty(prop_key="_PictType"),
    FormatterProperty(prop_key="_DurationNum"),
    FormatterProperty(prop_key="_DurationDen"),
    FormatterProperty(prop_key="_SARNum"),
    FormatterProperty(prop_key="_SARDen"),
]



from __future__ import annotations

from collections import deque
from collections.abc import Callable
from concurrent.futures import Future, wait
from logging import getLogger
from time import perf_counter_ns
from typing import TYPE_CHECKING

from jetpytools import clamp, cround
from PySide6.QtCore import QObject, Qt, QTime, QTimer, Signal, Slot
from PySide6.QtGui import QImage

from ...vsenv import run_in_background, run_in_loop
from ..outputs import AudioBuffer, FrameBuffer
from ..settings import SettingsManager
from ..views.timeline import Frame, Time, TimelineControlBar

if TYPE_CHECKING:
    from vsengine.policy import ManagedEnvironment

    from ...vsenv import QtEventLoop
    from ..outputs import OutputsManager
    from ..plugins.api import PluginAPI
    from .tab_manager import TabManager

logger = getLogger(__name__)

# Module constants
MIN_FRAME_DELAY_NS = 1_000_000  # 1ms minimum scheduling delay


class PlaybackState(QObject):
    """
    Manages playback-related state.

    Attributes:
        current_frame: Current frame being played.
        is_playing: Whether playback is currently active.

        last_fps_update_ns: Timestamp (ns) of last FPS display update.
        fps_history: Rolling window of frame timestamps for FPS averaging.

        buffer: Frame buffer for async pre-fetching during playback.
        frame_interval_ns: Target frame interval in nanoseconds for FPS limiting.
        next_frame_time_ns: Target time (ns) when next frame should start.

        audio_buffer: Audio buffer for async audio frame pre-fetching.
        audio_frame_interval_ns: Target audio frame interval in nanoseconds.
        next_audio_frame_time_ns: Target time (ns) when next audio frame should be pushed.

        video_timer: Timer for video frame scheduling.
        audio_timer: Timer for audio frame scheduling.

        _cleanup_future: Pending buffer cleanup future.
        _audio_cleanup_future: Pending audio buffer cleanup future.
    """

    def __init__(self, parent: QObject | None = None) -> None:
        super().__init__(parent)

        self.current_frame = 0
        self.is_playing = False

        self.last_fps_update_ns = 0
        self.fps_history = deque[int](maxlen=25)

        # Video playback state
        self.buffer: FrameBuffer | None = None
        self.frame_interval_ns = 0
        self.next_frame_time_ns = 0

        # Audio playback state
        self.audio_buffer: AudioBuffer | None = None
        self.audio_frame_interval_ns = 0
        self.next_audio_frame_time_ns = 0

        self.video_timer = QTimer(self, timerType=Qt.TimerType.PreciseTimer, singleShot=True)
        self.audio_timer = QTimer(self, timerType=Qt.TimerType.PreciseTimer, singleShot=True)

        self._cleanup_future: Future[None] | None = None
        self._audio_cleanup_future: Future[None] | None = None

    def reset(self) -> None:
        self.last_fps_update_ns = 0
        self.fps_history.clear()

        self.frame_interval_ns = 0
        self.next_frame_time_ns = 0

        self.audio_frame_interval_ns = 0
        self.next_audio_frame_time_ns = 0

        self.video_timer.stop()
        self.audio_timer.stop()

        if self.buffer:
            self._cleanup_future = self.buffer.invalidate()
            self.buffer = None

        if self.audio_buffer:
            self._audio_cleanup_future = self.audio_buffer.invalidate()
            self.audio_buffer = None

    def reset_audio(self) -> None:
        self.next_audio_frame_time_ns = 0
        self.audio_frame_interval_ns = 0
        self.audio_timer.stop()

        if self.audio_buffer:
            self._audio_cleanup_future = self.audio_buffer.invalidate()
            self.audio_buffer = None

    def wait_for_cleanup(self, timeout: float | None = None, stall_cb: Callable[[], None] | None = None) -> None:
        futures = list[Future[None]]()

        if self._cleanup_future:
            futures.append(self._cleanup_future)
        if self._audio_cleanup_future:
            futures.append(self._audio_cleanup_future)

        if futures:
            if timeout is not None and stall_cb:
                _, undone = wait(futures, timeout=timeout)
                if undone:
                    stall_cb()

            for f in futures:
                f.result()

        self._cleanup_future = None
        self._audio_cleanup_future = None


class PlaybackManager(QObject):
    """
    Manages video and audio playback.

    This class encapsulates all playback logic including buffering, frame timing,
    seeking, and audio synchronization. It communicates with the UI via signals.
    """

    # Signals for UI communication
    frameRendered = Signal(QImage, float)  # image, sar
    timelineCursorChanged = Signal(int)  # frame number

    audioOutputChanged = Signal(int)  # index

    statusLoadingStarted = Signal(str)  # message
    statusLoadingFinished = Signal(str)  # completed message
    statusLoadingErrored = Signal(str)  # error message

    loadFailed = Signal()  # error during frame render

    def __init__(
        self,
        loop: QtEventLoop,
        get_env: Callable[[], ManagedEnvironment],
        api: PluginAPI,
        outputs_manager: OutputsManager,
        tab_manager: TabManager,
        tbar: TimelineControlBar,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent)

        self._loop = loop
        self._get_env = get_env

        self._api = api
        self._outputs_manager = outputs_manager
        self._tab_manager = tab_manager
        self._tbar = tbar

        # Playback buttons signals
        self._tbar.playback_container.seek_n_back_btn.clicked.connect(lambda: self.seek_n_frames(-1))
        self._tbar.playback_container.seek_1_back_btn.clicked.connect(lambda: self.seek_frame(-1))
        self._tbar.playback_container.play_pause_btn.clicked.connect(self.toggle_playback)
        self._tbar.playback_container.seek_1_fwd_btn.clicked.connect(lambda: self.seek_frame(1))
        self._tbar.playback_container.seek_n_fwd_btn.clicked.connect(lambda: self.seek_n_frames(1))

        # Custom context menu signals
        self._tbar.playback_container.settingsChanged.connect(self._on_playback_settings_changed)
        self._tbar.playback_container.playZone.connect(self._on_play_zone)
        self._tbar.playback_container.audioDelayChanged.connect(self._on_audio_delay_changed)
        self._tbar.playback_container.audio_output_combo.currentIndexChanged.connect(self._on_audio_output_changed)

        # Connect Frame/Time Edit signals
        self._tbar.playback_container.frame_edit.frameChanged.connect(self._on_frame_changed)
        self._tbar.playback_container.time_edit.valueChanged.connect(self._on_time_changed)

        # Speaker and slider signals
        self._tbar.playback_container.volumeChanged.connect(self._on_volume_changed)
        self._tbar.playback_container.muteChanged.connect(self._on_mute_changed)

        # Timeline signal
        self._tbar.timeline.clicked.connect(self._on_timeline_clicked)

        # Playback state
        self.state = PlaybackState(self)
        self.state.video_timer.timeout.connect(self._play_next_frame)
        self.state.audio_timer.timeout.connect(self._play_next_audio_frame)

        self.can_reload = False
        self._timeline_rendering = False
        self._pending_frame: Frame | None = None

    @property
    def _env(self) -> ManagedEnvironment:
        return self._get_env()

    @run_in_loop
    def request_frame(self, n: int, cb_render: Callable[[Future[None]], None] | None = None) -> None:
        """Request a specific frame to be rendered and displayed."""
        logger.debug("Frame requested: %d", n)

        if not (voutput := self._outputs_manager.current_voutput):
            return

        n = clamp(n, 0, voutput.vs_output.clip.num_frames - 1)

        fut = self._render_frame(n)

        def on_complete(f: Future[None]) -> None:
            if f.exception():
                logger.error("Frame render failed", exc_info=f.exception())
                self.loadFailed.emit()
            elif self._tab_manager.tabs.currentIndex() != -1:
                voutput.last_frame = n
                self.state.current_frame = n

        fut.add_done_callback(on_complete)

        if cb_render:
            fut.add_done_callback(cb_render)

    @run_in_background(name="RenderFrame")
    def _render_frame(self, n: int) -> None:
        """Render a specific frame and emit for display."""
        logger.debug("Rendering frame %d (background)", n)

        if not (voutput := self._outputs_manager.current_voutput):
            return

        self.can_reload = False

        with self._env.use():
            if self.state.is_playing:
                with voutput.prepared_clip.get_frame(n) as frame:
                    logger.debug("Frame %d rendered", n)
                    image = voutput.packer.frame_to_qimage(frame)

                self._api._on_current_frame_changed(n, None)
            else:
                self.statusLoadingStarted.emit(f"Rendering frame {n}...")

                with voutput.prepared_clip.get_frame(n) as frame:
                    logger.debug("Frame %d rendered", n)
                    image = voutput.packer.frame_to_qimage(frame)

                self._api._on_current_frame_changed(n, None)
                self.statusLoadingFinished.emit("Completed")

        self.frameRendered.emit(image, self._get_sar_from_props(n))
        self.timelineCursorChanged.emit(n)
        self.can_reload = True

    @Slot(int)
    def seek_frame(self, delta: int) -> None:
        """Seek by a relative number of frames."""
        if not (voutput := self._outputs_manager.current_voutput):
            logger.warning("No current video output, ignoring seek")
            return

        self.request_frame(clamp(self.state.current_frame + delta, 0, voutput.vs_output.clip.num_frames - 1))

    @Slot(int)
    def seek_n_frames(self, direction: int) -> None:
        """Seek by N frames (configured step size)."""
        self.seek_frame(direction * self._tbar.playback_container.settings.seek_step)

    @Slot()
    def toggle_playback(self) -> None:
        """Toggle between play and pause."""
        self.state.is_playing = not self.state.is_playing

        if self.state.is_playing:
            self._start_playback()
        else:
            self._stop_playback()

    def stop(self) -> None:
        """Stop playback externally."""
        if self.state.is_playing:
            self._stop_playback()

    @run_in_background(name="StartPlayback")
    def _start_playback(self, play_range: range | None = None, loop: bool = False) -> None:
        logger.debug("Starting playback")

        if not (voutput := self._outputs_manager.current_voutput):
            return

        self._tbar.set_playback_controls_enabled(False)
        self._tbar.timeline.is_events_blocked = True
        self._loop.from_thread(self._tbar.playback_container.play_pause_btn.setChecked, True)
        self.can_reload = False

        try:
            # Wait for any pending buffer cleanup before creating new buffer
            # This prevents accumulation of buffers when user spams play/pause
            self.state.wait_for_cleanup(
                timeout=1.0,
                stall_cb=lambda: self.statusLoadingStarted.emit("Clearing buffer..."),
            )

            self.state.is_playing = True
            self.state.reset()

            if not play_range:
                play_range = range(self.state.current_frame, voutput.vs_output.clip.num_frames)

            self._prepare_video(play_range=play_range, loop=loop)
            self._prepare_audio(play_range=play_range, loop=loop)

            if self.state.audio_buffer:
                self.state.audio_buffer.wait_for_first_frame(
                    timeout=0.25,
                    stall_cb=lambda: self.statusLoadingStarted.emit("Buffering audio..."),
                )

            if self.state.buffer:
                self.state.buffer.wait_for_first_frame(
                    timeout=0.25,
                    stall_cb=lambda: self.statusLoadingStarted.emit("Buffering..."),
                )

            # Check if user cancelled
            if not self.state.is_playing:
                return

            self._api._on_playback_started()

            # Start both audio and video at the same time
            self.state.next_frame_time_ns = self.state.next_audio_frame_time_ns = perf_counter_ns()
            self.statusLoadingStarted.emit("Playing...")

            # Start the render chain. Each frame will chain to the next
            self._play_next_audio_frame()
            self._play_next_frame()
        except Exception:
            self._stop_playback()
            raise
        finally:
            self.can_reload = True

    @run_in_loop(return_future=False)
    def _stop_playback(self) -> None:
        logger.debug("Stopping playback")
        self._stop_audio()
        self.state.is_playing = False

        self.state.reset()

        self._tbar.set_playback_controls_enabled(True)
        self._tbar.timeline.is_events_blocked = False
        self._tbar.playback_container.play_pause_btn.setChecked(False)

        self.can_reload = True

        self._api._on_playback_stopped()
        self.statusLoadingFinished.emit("Paused")

    @run_in_loop(return_future=False)
    def _stop_audio(self) -> None:
        self.state.audio_timer.stop()

        if aoutput := self._outputs_manager.current_aoutput:
            aoutput.sink.reset()

        self.state.reset_audio()

    def _restart_playback(self) -> None:
        self._stop_playback()
        self.state.is_playing = True
        self._start_playback()

    @run_in_background(name="PlaybackNextFrame")
    def _play_next_frame(self) -> None:
        if not self.state.is_playing:
            return

        if not (voutput := self._outputs_manager.current_voutput):
            logger.error("No current video output during playback")
            self.toggle_playback()
            return

        if self.state.buffer:
            try:
                result = self.state.buffer.get_next_frame()
            except Exception as e:
                logger.error(
                    "An error occured during the rendering of the frame %d with the message: (%s): %s",
                    self.state.current_frame + 1,
                    e.__class__.__name__,
                    e,
                )
                self.loadFailed.emit()
                return

            if result:
                frame_n, frame, plugin_frames = result

                self._track_fps()

                self.state.current_frame = frame_n
                voutput.last_frame = frame_n

                try:
                    with self._env.use(), frame:
                        image = voutput.packer.frame_to_qimage(frame)

                    self.frameRendered.emit(image, self._get_sar_from_props(frame_n))
                    self.timelineCursorChanged.emit(frame_n)

                    self._api._on_current_frame_changed(frame_n, plugin_frames)
                finally:
                    for frame_to_close in plugin_frames.values():
                        frame_to_close.close()

                self._schedule_or_continue()
            else:
                self.toggle_playback()

    def _track_fps(self) -> None:
        now = perf_counter_ns()

        self.state.fps_history.append(now)

        if (total_elapsed := self.state.fps_history[-1] - self.state.fps_history[0]) > 0 and (
            now - self.state.last_fps_update_ns
            >= SettingsManager.global_settings.playback.fps_update_interval * 1_000_000_000
        ):
            self.statusLoadingStarted.emit(
                f"Playing @ {(len(self.state.fps_history) - 1) * 1_000_000_000 / total_elapsed:.3f} fps"
            )
            self.state.last_fps_update_ns = now

    def _schedule_or_continue(self) -> None:
        if self._tbar.playback_container.settings.uncapped:
            self._play_next_frame()
            return None

        if not self.state.is_playing:
            return None

        # VFR path
        if self.state.frame_interval_ns == 0 and (voutput := self._outputs_manager.current_voutput):
            # If framedurs has been provided
            if voutput.framedurs:
                self.state.next_frame_time_ns += cround(
                    1_000_000_000
                    * voutput.framedurs[self.state.current_frame]
                    / self._tbar.playback_container.settings.speed
                )
            # Fallback to frameprops for frame duration
            else:
                # Search from the end (most recent) for the closest available properties <= current frame
                props = next((voutput.props[n] for n in reversed(voutput.props) if n <= self.state.current_frame), None)

                if not props or not (dnum := props.get("_DurationNum")) or not (dden := props.get("_DurationDen")):
                    logger.warning("No duration props available for frame %d", self.state.current_frame)
                    return self._stop_playback()

                self.state.next_frame_time_ns += cround(
                    1_000_000_000 * dnum / (dden * self._tbar.playback_container.settings.speed)
                )
        else:
            self.state.next_frame_time_ns += self.state.frame_interval_ns

        # Schedule timer only if delay is significant; avoids timer overhead for <1ms delays
        if (delay_ns := self.state.next_frame_time_ns - perf_counter_ns()) > MIN_FRAME_DELAY_NS:
            self._loop.from_thread(lambda: self.state.video_timer.start(cround(delay_ns / 1_000_000)))
        else:
            self._play_next_frame()
        return None

    def _prepare_video(self, play_range: range, loop: bool = False) -> None:
        if not (voutput := self._outputs_manager.current_voutput):
            raise RuntimeError("No video output available")

        # Calculate target frame interval for FPS limiting
        if self._tbar.playback_container.settings.uncapped:
            self.state.frame_interval_ns = 0
        elif voutput.vs_output.clip.fps > 0:
            self.state.frame_interval_ns = cround(
                1_000_000_000
                * voutput.vs_output.clip.fps.denominator
                / (voutput.vs_output.clip.fps.numerator * self._tbar.playback_container.settings.speed)
            )
        else:
            logger.debug("VFR detected")
            self.state.frame_interval_ns = 0

            if not voutput.framedurs:
                props = voutput.props[self.state.current_frame]

                if "_DurationNum" not in props or "_DurationDen" not in props:
                    raise RuntimeError(
                        "Both '_DurationNum' and '_DurationDen' props need to be available for VFR playback"
                    )

        # Create and allocate video buffer
        self.state.buffer = FrameBuffer(video_output=voutput, env=self._env)
        self._api._register_plugin_nodes_to_buffer()

        self.state.buffer.allocate(play_range=play_range, loop=loop)

        logger.debug(
            "Target frame interval: %d ns (fps=%s), buffer_size=%d",
            self.state.frame_interval_ns,
            voutput.vs_output.clip.fps,
            self.state.buffer._size,
        )

    def _prepare_audio(self, play_range: range, loop: bool = False) -> None:
        if (
            self._tbar.playback_container.is_muted
            or not (aoutput := self._outputs_manager.current_aoutput)
            or not (voutput := self._outputs_manager.current_voutput)
        ):
            return

        if self._tbar.playback_container.settings.uncapped:
            logger.info("Uncapped settings detected, no audio will be played")
            return

        if play_range.step != 1:
            logger.info("Audio skipped due to non-standard step")
            return

        if not aoutput.setup_sink(
            self._tbar.playback_container.settings.speed,
            self._tbar.playback_container.volume,
        ):
            return

        # Audio can only be played reliably if there are frame durations
        if not voutput.cum_durations:
            logger.warning("No frame durations available, audio will not be played")
            return

        with self._env.use():
            # Audio prepared from play_range.start + 1 to match video which skips its first frame
            aoutput.prepare_playback_audio(
                voutput.frame_to_time(play_range.start + 1).total_seconds(),
                voutput.frame_to_time(play_range.stop).total_seconds(),
            )

        self.state.audio_buffer = AudioBuffer(aoutput, self._env)
        self.state.audio_buffer.allocate(play_range=range(aoutput.playback_audio.num_frames), loop=loop)

        # Calculate interval so that (num_frames x interval) = actual audio duration
        # This ensures timer and audio content stay in sync, especially for partial last frames
        audio_duration_ns = cround(
            1_000_000_000 * aoutput.playback_audio.num_samples / aoutput.prepared_audio.sample_rate
        )
        self.state.audio_frame_interval_ns = cround(
            audio_duration_ns / (aoutput.playback_audio.num_frames * self._tbar.playback_container.settings.speed)
        )

        logger.debug("Audio prepared: interval=%d ns", self.state.audio_frame_interval_ns)

    @run_in_loop
    def _play_next_audio_frame(self) -> None:
        if (
            not self.state.is_playing
            or not (aoutput := self._outputs_manager.current_aoutput)
            or self._tbar.playback_container.is_muted
            or self.state.audio_frame_interval_ns <= 0
        ):
            self.state.audio_timer.stop()
            return

        # Use absolute targets to prevent clock drift; catch up immediately if lagging
        self.state.next_audio_frame_time_ns += self.state.audio_frame_interval_ns

        if (
            aoutput.sink.bytesFree() >= aoutput.bytes_per_frame
            and self.state.audio_buffer
            and (result := self.state.audio_buffer.get_next_frame())
        ):
            with (frame := result[1]):
                aoutput.render_raw_audio_frame(frame)

        # Audio uses strict 0 to minimize jitter; any lag must be processed immediately
        if (delay_ns := self.state.next_audio_frame_time_ns - perf_counter_ns()) > 0:
            self.state.audio_timer.start(cround(delay_ns / 1_000_000))
        else:
            self._play_next_audio_frame()

    def _get_sar_from_props(self, n: int) -> float:
        if TYPE_CHECKING:
            assert self._outputs_manager.current_voutput

        props = self._outputs_manager.current_voutput.props.get(n)

        if not props:
            return 1.0

        sar_num, sar_den = props.get("_SARNum"), props.get("_SARDen")

        if isinstance(sar_num, int) and isinstance(sar_den, int):
            return sar_num / sar_den

        return 1.0

    # Signals
    @Slot(int, float, bool)
    def _on_playback_settings_changed(self, seek_step: int, speed: float, uncapped: bool) -> None:
        """Handle playback settings change from UI."""
        self._tbar.playback_container.settings.seek_step = seek_step
        self._tbar.playback_container.settings.speed = speed
        self._tbar.playback_container.settings.uncapped = uncapped

        if self.state.is_playing:
            self._restart_playback()

    def _on_play_zone(self, zone_frames: int, loop: bool, step: int) -> None:
        """Play a specific zone of frames."""
        if not (voutput := self._outputs_manager.current_voutput):
            return

        direction = 1 if step > 0 else -1

        end_frame = clamp(
            self.state.current_frame + (zone_frames * direction),
            0,
            voutput.vs_output.clip.num_frames - 1,
        )

        play_range = range(self.state.current_frame, end_frame + direction, step)

        self._tbar.playback_container.play_pause_btn.setChecked(True)
        self._start_playback(play_range=play_range, loop=loop)

    def _on_audio_delay_changed(self, delay_s: float) -> None:
        """Handle audio delay change from UI."""
        if self.state.is_playing:
            self._stop_playback()

        with self._env.use():
            for aoutput in self._outputs_manager.aoutputs:
                aoutput.prepare_audio(delay_s, self._api)

    def _on_audio_output_changed(self, index: int) -> None:
        """Handle audio output selection change."""
        self.audioOutputChanged.emit(index)

        if self.state.is_playing:
            self._restart_playback()

    @Slot(Frame, Frame)
    def _on_frame_changed(self, frame: Frame, old_frame: Frame) -> None:
        logger.debug("Frame changed: frame=%d", frame)
        self.request_frame(frame)

    @Slot(float)
    def _on_volume_changed(self, volume: float) -> None:
        if aoutput := self._outputs_manager.current_aoutput:
            aoutput.volume = volume

    def _on_mute_changed(self, is_muted: bool) -> None:
        if is_muted:
            self._stop_audio()
        elif self.state.is_playing:
            self._restart_playback()

    @Slot(QTime, QTime)
    def _on_time_changed(self, time: QTime, old_time: QTime) -> None:
        logger.debug("Time changed: time=%s", time)

        if not (voutput := self._outputs_manager.current_voutput):
            return

        frame = voutput.time_to_frame(Time.from_qtime(time))
        logger.debug("Time changed: frame=%d", frame)

        self.request_frame(frame)

    @Slot(Frame, Time)
    def _on_timeline_clicked(self, frame: Frame, time: Time) -> None:
        self.stop()

        self._pending_frame = frame

        if not self._timeline_rendering:
            logger.debug("Timeline clicked: frame=%d, time=%s", frame, time)
            self._render_pending_frame()

    def _render_pending_frame(self) -> None:
        if self._pending_frame is None:
            self._timeline_rendering = False
            return

        frame = self._pending_frame

        self._pending_frame = None
        self._timeline_rendering = True

        @run_in_loop(return_future=False)
        def on_render_complete(f: Future[None]) -> None:
            self._render_pending_frame()

        self.request_frame(frame, cb_render=on_render_complete)



import pathlib
import sys

import qdarkstyle  # type: ignore
from PySide6.QtWidgets import QApplication
from qdarkstyle.dark.palette import DarkPalette  # type: ignore
from qdarkstyle.light.palette import LightPalette  # type: ignore

app = QApplication(sys.argv)

# Dump dark theme
dark_stylesheet = qdarkstyle.load_stylesheet(palette=DarkPalette)
pathlib.Path("qdarkstyle_dark.qss").write_text(dark_stylesheet)
print("Saved: qdarkstyle_dark.qss")

# Dump light theme
light_stylesheet = qdarkstyle.load_stylesheet(palette=LightPalette)
pathlib.Path("qdarkstyle_light.qss").write_text(light_stylesheet)
print("Saved: qdarkstyle_light.qss")



from __future__ import annotations

from array import array
from contextlib import suppress
from enum import Enum, auto
from fractions import Fraction
from logging import getLogger
from math import ceil
from typing import TYPE_CHECKING, Any

import vapoursynth as vs
from jetpytools import clamp, cround
from PySide6.QtCore import QIODevice
from PySide6.QtMultimedia import QAudio, QAudioFormat, QAudioSink
from PySide6.QtWidgets import QApplication

from ...vsenv import run_in_loop
from ..plugins.manager import PluginManager
from ..settings import SettingsManager
from ..utils import LRUCache

if TYPE_CHECKING:
    from ...api._helpers import AudioMetadata
    from ..plugins import PluginAPI
    from ..views.timeline import Frame, Time

logger = getLogger(__name__)


class PrettyChannelsLayout(Enum):
    UNKNOWN = auto(), "Unknown"
    STEREO = (
        (
            vs.AudioChannels.FRONT_LEFT,
            vs.AudioChannels.FRONT_RIGHT,
        ),
        "2.0",
    )
    SURROUND_5_1 = (
        (
            vs.AudioChannels.FRONT_LEFT,
            vs.AudioChannels.FRONT_RIGHT,
            vs.AudioChannels.FRONT_CENTER,
            vs.AudioChannels.LOW_FREQUENCY,
            vs.AudioChannels.SIDE_LEFT,
            vs.AudioChannels.SIDE_RIGHT,
        ),
        "5.1",
    )
    SURROUND_7_1 = (
        (
            vs.AudioChannels.FRONT_LEFT,
            vs.AudioChannels.FRONT_RIGHT,
            vs.AudioChannels.FRONT_CENTER,
            vs.AudioChannels.LOW_FREQUENCY,
            vs.AudioChannels.BACK_LEFT,
            vs.AudioChannels.BACK_RIGHT,
            vs.AudioChannels.SIDE_LEFT,
            vs.AudioChannels.SIDE_RIGHT,
        ),
        "7.1",
    )

    pretty_name: str

    def __init__(self, value: Any, pretty_name: str = "") -> None:
        self._value_ = value
        self.pretty_name = pretty_name

    @classmethod
    def _missing_(cls, value: object) -> Any:
        return next((member for member in cls if member.value == value), cls.UNKNOWN)


class AudioOutput:
    # Standard downmix coefficient for center/surround channels: sqrt(2)/2
    DOWNMIX_COEFF = 0.7071067811865476
    # https://github.com/vapoursynth/vapoursynth/blob/8cd1cba539bf70eea21dc242d43349603115632d/include/VapourSynth4.h#L36
    SAMPLES_PER_FRAME = 3072

    def __init__(self, vs_output: vs.AudioNode, vs_index: int, metadata: AudioMetadata | None) -> None:
        self.vs_output = vs_output
        self.vs_index = vs_index
        self.vs_name = metadata.name if metadata else f"Audio {vs_index}"  # Matches vsview.set_output
        self.downmix = (
            metadata.downmix
            if metadata and metadata.downmix is not None
            else SettingsManager.global_settings.playback.downmix
        )
        self.chanels_layout = PrettyChannelsLayout(tuple(self.vs_output.channels))

        self._cache_delay_audio = LRUCache[float, vs.AudioNode]()

    @property
    def fps(self) -> Fraction:
        return Fraction(self.prepared_audio.sample_rate, self.SAMPLES_PER_FRAME)

    @property
    def bytes_per_frame(self) -> int:
        return self.SAMPLES_PER_FRAME * self.prepared_audio.num_channels * self.prepared_audio.bytes_per_sample

    @property
    def volume(self) -> float:
        return getattr(self, "_volume", 0.5)

    @volume.setter
    def volume(self, value: float) -> None:
        self._volume = clamp(value, 0.0, 1.0)
        self.sink.setVolume(self._volume)

    def prepare_audio(self, delay_s: float, api: PluginAPI) -> None:
        audio = self._cache_delay_audio.get(delay_s)

        if audio is None:
            audio = self.vs_output

            if PluginManager.audio_processor:
                audio = PluginManager.audio_processor(api).prepare(audio)

            if self.vs_output.num_channels > 2 and self.downmix:
                audio = self.create_stereo_downmix(audio)

            # Apply audio delay
            if delay_s != 0:
                delay_samples = round(delay_s * audio.sample_rate)
                abs_delay = abs(delay_samples)
                silence = audio.std.BlankAudio(length=abs_delay, keep=True)

                if delay_samples > 0:
                    audio = (silence + audio)[: audio.num_samples]
                elif delay_samples < 0:
                    audio = (audio[abs_delay:] + silence)[: audio.num_samples]

            self._cache_delay_audio[delay_s] = audio

        self.prepared_audio = audio

        # Playback node
        if audio.sample_type == vs.FLOAT:
            self.array_type = "f"
            sample_format = QAudioFormat.SampleFormat.Float
        elif audio.bits_per_sample <= 16:
            self.array_type = "h"
            sample_format = QAudioFormat.SampleFormat.Int16
        else:
            self.array_type = "i"
            sample_format = QAudioFormat.SampleFormat.Int32

        self._audio_buffer = array(self.array_type, [0] * (self.SAMPLES_PER_FRAME * audio.num_channels))

        self.qformat = QAudioFormat()
        self.qformat.setChannelCount(audio.num_channels)
        self.qformat.setSampleRate(audio.sample_rate)
        self.qformat.setSampleFormat(sample_format)

        self.sink = AudioSink(self.qformat)

    def prepare_playback_audio(self, start_seconds: float, stop_seconds: float) -> None:
        duration_samples = ceil((stop_seconds - start_seconds) * self.prepared_audio.sample_rate)

        # Truncate for start to avoid cutting into the first sample
        start = max(0, int(start_seconds * self.prepared_audio.sample_rate))
        end = min(self.prepared_audio.num_samples, start + duration_samples)

        self.playback_audio = self.prepared_audio[start:end]
        self.playback_audio.std.SetAudioCache(0)

    def clear(self) -> None:
        """Clear VapourSynth resources."""
        if hasattr(self, "sink"):
            self.sink.stop()
            self.sink.deleteLater()
            del self.sink

        for name in ["prepared_audio", "vs_output", "playback_audio"]:
            with suppress(AttributeError):
                delattr(self, name)

        self._cache_delay_audio.clear()

    def render_raw_audio_frame(self, frame: vs.AudioFrame) -> None:
        if not self.sink.ready:
            return

        num_channels = self.prepared_audio.num_channels

        if num_channels == 1:
            data = frame[0].tobytes()
        else:
            # We can get the number of samples from the shape of the first channel
            required_len = frame[0].shape[0] * num_channels

            buffer = (
                self._audio_buffer
                if required_len == len(self._audio_buffer)
                else array(self.array_type, [0] * required_len)
            )

            for i in range(num_channels):
                buffer[i::num_channels] = array(self.array_type, frame[i].tobytes())

            data = buffer.tobytes()

        # Handle cases where the sink is stopped between the initial check and writing
        with suppress(RuntimeError):
            self.sink.io.write(data)

    @run_in_loop(return_future=False)
    def setup_sink(self, speed: float = 1.0, volume: float = 1.0) -> bool:
        """
        Initialize the audio sink for playback.

        Returns True if successful, False if audio format not supported.
        """
        self.qformat.setSampleRate(round(self.prepared_audio.sample_rate * speed))
        self.sink.stop()
        self.sink.deleteLater()
        self.sink = AudioSink(self.qformat)
        self.sink.setup(SettingsManager.global_settings.playback.audio_buffer_size * self.bytes_per_frame, self.volume)

        if not self.sink.ready:
            logger.error(
                "Failed to start audio sink - format may not be supported (%d Hz, speed=%.2fx)",
                self.qformat.sampleRate(),
                speed,
            )
            return False

        self.volume = volume

        logger.debug(
            "Audio sink initialized: %d Hz (base %d Hz, speed=%.2fx), %d channels, format=%s",
            self.qformat.sampleRate(),
            self.prepared_audio.sample_rate,
            speed,
            self.prepared_audio.num_channels,
            self.qformat.sampleFormat(),
        )
        return True

    def create_stereo_downmix(self, audio: vs.AudioNode) -> vs.AudioNode:
        """
        Create a stereo downmix of the source audio using std.AudioMix.
        """

        # Build downmix matrix
        # 5.1/7.1 to stereo downmix coefficients:
        # L = 1.0*L + 0.707*C + 0.707*Ls + 0.707*Lb
        # R = 1.0*R + 0.707*C + 0.707*Rs + 0.707*Rb

        left_coeffs = [0.0] * audio.num_channels
        right_coeffs = [0.0] * audio.num_channels

        for i, channel in enumerate(audio.channels):
            match channel:
                case vs.AudioChannels.FRONT_LEFT:
                    left_coeffs[i] = 1.0
                case vs.AudioChannels.FRONT_RIGHT:
                    right_coeffs[i] = 1.0
                case vs.AudioChannels.FRONT_CENTER:
                    left_coeffs[i] = self.DOWNMIX_COEFF
                    right_coeffs[i] = self.DOWNMIX_COEFF
                case vs.AudioChannels.SIDE_LEFT | vs.AudioChannels.BACK_LEFT:
                    left_coeffs[i] = self.DOWNMIX_COEFF
                case vs.AudioChannels.SIDE_RIGHT | vs.AudioChannels.BACK_RIGHT:
                    right_coeffs[i] = self.DOWNMIX_COEFF

        normalization = max(sum(c**2 for c in left_coeffs) ** 0.5, sum(c**2 for c in right_coeffs) ** 0.5, 1.0)

        final_matrix = [(c / normalization) for c in (left_coeffs + right_coeffs)]

        logger.debug(
            "Creating downmix: %d channels -> Stereo. Normalization: %.4f. Matrix: %s. Layout: %s",
            audio.num_channels,
            normalization,
            final_matrix,
            [c.name for c in audio.channels],
        )

        return audio.std.AudioMix(matrix=final_matrix, channels_out=[vs.FRONT_LEFT, vs.FRONT_RIGHT])

    def time_to_frame(self, seconds: float, *, eps: float = 1e-6) -> Frame:
        from ..views.timeline import Frame

        return Frame(cround(seconds * self.fps, eps=eps))

    def frame_to_time(self, frame_num: int) -> Time:
        from ..views.timeline import Time

        return Time(seconds=float(frame_num / self.fps))

    def frame_to_sample(self, frame_num: int) -> int:
        return frame_num * self.SAMPLES_PER_FRAME

    def sample_to_frame(self, sample_num: int, *, eps: float = 1e-6) -> Frame:
        from ..views.timeline import Frame

        return Frame(cround(sample_num / self.SAMPLES_PER_FRAME, eps=eps))

    def time_to_sample(self, seconds: float, *, eps: float = 1e-6) -> int:
        return cround(seconds * self.prepared_audio.sample_rate, eps=eps)

    def sample_to_time(self, sample_num: int) -> Time:
        from ..views.timeline import Time

        return Time(seconds=sample_num / self.prepared_audio.sample_rate)

    def __del__(self) -> None:
        self.clear()


class AudioSink(QAudioSink):
    def __init__(self, format: QAudioFormat) -> None:
        super().__init__(format)
        self.ready = False
        # Move the sink to the main thread so it can be controlled from LoaderWorkspace._stop_audio
        self.moveToThread(QApplication.instance().thread())  # type: ignore[union-attr]

    def setVolume(self, volume: float) -> None:
        """Override to handle perceptual to linear volume conversion."""
        super().setVolume(
            QAudio.convertVolume(
                volume,
                QAudio.VolumeScale.LogarithmicVolumeScale,
                QAudio.VolumeScale.LinearVolumeScale,
            )
        )

    @property
    def io(self) -> QIODevice:
        if not self.ready:
            raise RuntimeError("Device not ready")
        return self._iodevice

    def stop(self) -> None:
        self.ready = False
        super().stop()

    def reset(self) -> None:
        self.ready = False
        super().reset()

    def setup(self, buffer_size: int, volume: float) -> None:
        self.setBufferSize(buffer_size)
        self.setVolume(volume)

        self._iodevice = self.start()

        if not self._iodevice:
            return

        self.ready = True



from .audio import AudioOutput
from .buffer import AudioBuffer, FrameBuffer
from .manager import OutputsManager
from .packing import Packer, get_packer
from .video import VideoOutput

__all__ = ["AudioBuffer", "AudioOutput", "FrameBuffer", "OutputsManager", "Packer", "VideoOutput", "get_packer"]



from __future__ import annotations

import pkgutil
import weakref
from collections.abc import Callable
from concurrent.futures import Future
from functools import wraps
from importlib import import_module
from inspect import ismethod
from logging import getLogger
from pathlib import Path
from threading import Lock
from typing import TYPE_CHECKING, Any, Concatenate, Literal

import pluggy
from jetpytools import Singleton, inject_self
from pydantic import BaseModel, ValidationError
from vapoursynth import AudioNode, VideoNode
from vsengine.loops import EventLoop, get_loop

from ...vsenv import run_in_background
from .. import tools
from . import specs

if TYPE_CHECKING:
    from .api import NodeProcessor, WidgetPluginBase

logger = getLogger(__name__)


class Notifier:
    def __init__(self) -> None:
        self._callbacks = list[tuple[weakref.ReferenceType[Callable[[], Any]], EventLoop]]()
        self._ready = False
        self._lock = Lock()

    def register(self, callback: Callable[[], Any]) -> None:
        with self._lock:
            if self._ready:
                get_loop().from_thread(callback)
                return

            self._callbacks.append(
                (weakref.WeakMethod(callback) if ismethod(callback) else weakref.ref(callback), get_loop())
            )

    def notify(self) -> None:
        with self._lock:
            if self._ready:
                return

            self._ready = True
            current_callbacks = self._callbacks[:]
            self._callbacks.clear()

        for ref, loop in current_callbacks:
            if (cb := ref()) is not None:
                loop.from_thread(cb)


def ensure_loaded[T: PluginManager, **P, R](
    action: str,
) -> Callable[[Callable[Concatenate[T, P], R]], Callable[Concatenate[T, P], R]]:

    def decorator(func: Callable[Concatenate[T, P], R]) -> Callable[Concatenate[T, P], R]:
        @wraps(func)
        def wrapper(self: T, *args: P.args, **kwargs: P.kwargs) -> R:
            if action == "wait" and not self.loaded:
                self.load()
                self.wait_for_loaded()
            elif action == "entrypoints" and not self._entry_points_loaded:
                raise RuntimeError("PluginManager is not loaded yet")

            return func(self, *args, **kwargs)

        return wrapper

    return decorator


class PluginManager(Singleton):
    def __init__(self) -> None:
        self.manager = pluggy.PluginManager("vsview")
        self._notifier = Notifier()
        self._settings_extracted = False
        self._entry_points_loaded = False
        self._load_future: Future[None] | None = None
        self._lock = Lock()

    @inject_self.cached.property
    @ensure_loaded("entrypoints")
    def tooldocks(self) -> list[type[WidgetPluginBase]]:
        return self.manager.hook.vsview_register_tooldock()

    @inject_self.cached.property
    @ensure_loaded("entrypoints")
    def toolpanels(self) -> list[type[WidgetPluginBase]]:
        return self.manager.hook.vsview_register_toolpanel()

    @inject_self.cached.property
    @ensure_loaded("entrypoints")
    def video_processor(self) -> type[NodeProcessor[VideoNode]] | None:
        return self.manager.hook.vsview_get_video_processor()

    @inject_self.cached.property
    @ensure_loaded("entrypoints")
    def audio_processor(self) -> type[NodeProcessor[AudioNode]] | None:
        return self.manager.hook.vsview_get_audio_processor()

    @inject_self.property
    @ensure_loaded("entrypoints")
    def all_plugins(self) -> list[type[WidgetPluginBase | NodeProcessor[Any]]]:
        all_plugins: set[Any] = {*self.tooldocks, *self.toolpanels}

        if vp := self.video_processor:
            all_plugins.add(vp)
        if ap := self.audio_processor:
            all_plugins.add(ap)

        return sorted(all_plugins, key=lambda p: p.identifier)

    @inject_self.property
    def loaded(self) -> bool:
        return self._load_future is not None and self._load_future.done()

    @inject_self
    def load(self) -> None:
        if self._load_future:
            return

        self._load_future = self._load_worker()

    @inject_self
    def wait_for_loaded(self) -> None:
        if self._load_future:
            self._load_future.result()

    @inject_self
    def call_when_loaded(self, cb: Callable[[], Any]) -> None:
        self._notifier.register(cb)

    @run_in_background(name="PluginManagerLoad")
    def _load_worker(self) -> None:
        self.manager.add_hookspecs(specs)

        for _, name, _ in pkgutil.iter_modules(tools.__path__):
            if name.startswith("_"):
                continue
            logger.debug("Registering %s first party plugin", name)
            self.manager.register(import_module(f"vsview.app.tools.{name}"))

        logger.debug("Loading entrypoints...")
        n = self.manager.load_setuptools_entrypoints("vsview")
        logger.debug("Loaded %d second/third party plugins", n)
        self._entry_points_loaded = True

        self._register_shortcuts()
        self._construct_settings_registry()

        logger.debug("Plugin integration, finalized")

        # Fire signal
        self._notifier.notify()

    def _register_shortcuts(self) -> None:
        from ..settings.shortcuts import ShortcutManager

        for plugin in self.all_plugins:
            if not (shortcuts := getattr(plugin, "shortcuts", ())):
                continue

            expected_prefix = f"{plugin.identifier}."
            valid_definitions = []

            for definition in shortcuts:
                if not definition.startswith(expected_prefix):
                    logger.warning(
                        "Plugin %r has shortcut %r without proper namespace prefix. "
                        "Expected prefix: %r. Shortcut will be ignored.",
                        plugin.identifier,
                        str(definition),
                        expected_prefix,
                    )
                    continue
                valid_definitions.append(definition)

            if valid_definitions:
                ShortcutManager.register_definitions(valid_definitions)

        # FIXME
        # ShortcutManager._check_conflicts()

    def _construct_settings_registry(self) -> None:
        from ..settings.dialog import SettingsDialog
        from ..settings.models import SettingEntry, extract_settings

        def extract_plugin_settings(model: type, plugin_id: str, section_name: str) -> list[SettingEntry]:
            return [
                entry._replace(key=f"plugins.{plugin_id}.{entry.key}")
                for entry in extract_settings(model, section=section_name)
            ]

        global_entries = list[SettingEntry]()
        local_entries = list[SettingEntry]()

        for plugin in self.all_plugins:
            global_model = plugin.global_settings_model
            local_model = plugin.local_settings_model
            identifier = plugin.identifier
            display_name = plugin.display_name

            if global_model is None and local_model is None:
                continue

            section = f"Plugin - {display_name}"

            if global_model is not None:
                global_entries.extend(extract_plugin_settings(global_model, identifier, section))

            if local_model is not None:
                local_entries.extend(extract_plugin_settings(local_model, identifier, section))

        self._populate_default_settings("global")

        # Extend dialog registries
        SettingsDialog.global_settings_registry.extend(global_entries)
        SettingsDialog.local_settings_registry.extend(local_entries)

        logger.debug("Plugin settings extracted")

    @inject_self
    @ensure_loaded("wait")
    def populate_default_settings(self, scope: Literal["global", "local"], file_path: Path | None = None) -> None:
        self._populate_default_settings(scope, file_path)

    def _populate_default_settings(self, scope: Literal["global", "local"], file_path: Path | None = None) -> None:
        from ..settings import SettingsManager

        if scope == "local" and file_path is not None:
            settings_container = SettingsManager.get_local_settings(file_path)
        else:
            settings_container = SettingsManager.global_settings

        model_attr = f"{scope}_settings_model"

        with self._lock:
            for plugin in self.all_plugins:
                if (model := getattr(plugin, model_attr)) is None:
                    continue

                raw = settings_container.plugins.get(plugin.identifier, {})
                existing = raw.model_dump() if isinstance(raw, BaseModel) else raw

                # Validate existing settings (missing fields will be filled with defaults by Pydantic)
                try:
                    validaded = model.model_validate(existing)
                except ValidationError:
                    logger.exception("The plugin %r has invalid settings:", plugin.identifier)
                    continue

                settings_container.plugins[plugin.identifier] = validaded



from pathlib import Path
from typing import Any

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CustomBuildHook(BuildHookInterface[Any]):
    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        sbom_path = Path(self.root) / "sboms.json"
        if sbom_path.exists():
            build_data.setdefault("sbom_files", []).append(str(sbom_path.relative_to(self.root)))



import ast
import json
import re
from bisect import bisect_left
from datetime import timedelta
from fractions import Fraction
from logging import getLogger
from math import ceil
from typing import Any, BinaryIO

from jetpytools import fallback

from .api import Parser, borrowed_text_wrapper
from .models import RangeFrame, RangeTime, SceneRow

logger = getLogger(__name__)


class AssParser(Parser):
    filter = Parser.FileFilter("Aegisub Advanced SSA subtitles", "ass")

    def parse(self, io: BinaryIO, name: str, fps: Fraction) -> SceneRow:
        with borrowed_text_wrapper(io, encoding="utf-8-sig") as wrapper:
            text = wrapper.read()

        ranges = list[RangeFrame]()

        for matched in re.finditer(r"^Dialogue:\s\d+,([^,]+),([^,]+),(?:[^,]*,){6}(.*)$", text, re.MULTILINE):
            start_ts, end_ts, txt = matched.groups()

            h, m, s = start_ts.split(":")
            start_seconds = int(h) * 3600 + int(m) * 60 + float(s)

            h, m, s = end_ts.split(":")
            end_seconds = int(h) * 3600 + int(m) * 60 + float(s)

            # formula is from videotimestamps with a rounding method of "round"
            # https://github.com/moi15moi/VideoTimestamps/blob/9d8259a94d069d7f85d6ab502b6ded3bfb25145a/video_timestamps/fps_timestamps.py#L65-L85
            start_frame = ceil(((ceil(start_seconds * 1000) - 0.5) / 1000) * fps + 1) - 1
            end_frame = ceil(((ceil(end_seconds * 1000) - 0.5) / 1000) * fps) - 1

            ranges.append(RangeFrame(start=start_frame, end=end_frame, label=txt))

        return SceneRow(color=self.get_color(), name=name, ranges=ranges)


class OGMParser(Parser):
    filter = Parser.FileFilter("Ogg Media (OGM) Chapters", "txt")

    def parse(self, io: BinaryIO, name: str, fps: Fraction) -> SceneRow:
        pattern = re.compile(
            r"^\s*(CHAPTER\d+)\s*=\s*(\d+):(\d+):(\d+(?:\.\d+)?)\s*[\r\n]+"
            r"\s*\1NAME\s*=\s*(.*)",
            re.MULTILINE,
        )

        ranges = list[RangeTime]()

        with borrowed_text_wrapper(io) as wrapper:
            text = wrapper.read()

        if not text:
            raise ValueError("Empty file")

        for match in pattern.finditer(text):
            hours, minutes, seconds_str, name_val = match.group(2, 3, 4, 5)

            ranges.append(
                RangeTime(
                    start=timedelta(hours=int(hours), minutes=int(minutes), seconds=float(seconds_str)),
                    label=name_val.strip(),
                )
            )

        return SceneRow(color=self.get_color(), name=name, ranges=ranges)


class MatroskaXMLParser(Parser):
    filter = Parser.FileFilter("Matroska XML Chapters", "xml")

    def parse(self, io: BinaryIO, name: str, fps: Fraction) -> SceneRow | list[SceneRow]:
        import xml.etree.ElementTree as ET

        try:
            tree = ET.parse(io)
            root = tree.getroot()
        except ET.ParseError:
            raise ValueError(f"Could not parse XML file: {name}")

        editions = root.findall(".//EditionEntry")

        # Fallback if no EditionEntry is found
        if not editions:
            return SceneRow(color=self.get_color(), name=name, ranges=self._parse_atoms(root))

        scenes = list[SceneRow]()
        for i, edition in enumerate(editions, 1):
            if not (ranges := self._parse_atoms(edition)):
                continue

            uid_node = edition.find("EditionUID")
            uid_str = f" ({uid_node.text})" if uid_node is not None and uid_node.text else f" (Edition {i})"

            # If only one edition, keep the simple name
            name_edition = name if len(editions) == 1 else f"{name}{uid_str}"
            scenes.append(SceneRow(color=self.get_color(), name=name_edition, ranges=ranges))

        return scenes

    def _parse_atoms(self, parent: Any) -> list[RangeTime]:
        ranges = list[RangeTime]()

        # Find all ChapterAtom nodes
        for atom in parent.findall(".//ChapterAtom"):
            time_node = atom.find("ChapterTimeStart")

            if time_node is None or time_node.text is None:
                continue

            try:
                # Format is HH:MM:SS.nnnnnnnnn
                # We split manually because datetime.strptime chokes on 9-digit nanoseconds
                parts = time_node.text.split(":")
                if len(parts) == 3:
                    start_dt = timedelta(hours=int(parts[0]), minutes=int(parts[1]), seconds=float(parts[2]))

                    label = ""
                    if (display_node := atom.find("ChapterDisplay")) is not None:
                        name_node = display_node.find("ChapterString")

                        if name_node is not None and name_node.text:
                            label = name_node.text

                    ranges.append(RangeTime(start=start_dt, label=label))
            except ValueError:
                continue

        return ranges


class XvidLogParser(Parser):
    filter = Parser.FileFilter("XviD Log", ["txt", "log"])

    def parse(self, io: BinaryIO, name: str, fps: Fraction) -> SceneRow:
        ranges = list[RangeFrame]()
        current_frame = 0

        with borrowed_text_wrapper(io, encoding="utf-8") as wrapper:
            lines = wrapper.read()

        for line in lines.splitlines():
            line = line.strip()

            if not line or line.startswith("#"):
                continue

            if line[0] == "i":
                ranges.append(RangeFrame(start=current_frame))

            current_frame += 1

        return SceneRow(color=self.get_color(), name=name, ranges=ranges)


class QPFileParser(Parser):
    filter = Parser.FileFilter("QP File", ["qp", "txt"])

    def parse(self, io: BinaryIO, name: str, fps: Fraction) -> SceneRow:
        ranges = list[RangeFrame]()

        with borrowed_text_wrapper(io, encoding="utf-8") as wrapper:
            text = wrapper.read()

        for matched in re.finditer(r"^(\d+)\s+[IK]\s+(-?\d+)", text, re.MULTILINE):
            try:
                frame = int(matched.group(1))
            except ValueError:
                continue

            ranges.append(RangeFrame(start=frame, label=str(matched.group(2))))

        return SceneRow(color=self.get_color(), name=name, ranges=ranges)


class WobblyParser(Parser):
    filter = Parser.FileFilter("Wobbly File", "wob")

    def parse(self, io: BinaryIO, name: str, fps: Fraction) -> SceneRow | list[SceneRow]:
        with borrowed_text_wrapper(io, encoding="utf-8") as wrapper:
            text = wrapper.read()

        try:
            data: dict[str, Any] = json.loads(text)
        except Exception as e:
            raise ValueError(f"Could not parse Wobbly file: {name}") from e

        decimations = sorted(data.get("decimated frames", []))
        scenes = list[SceneRow]()

        # Sections
        if sections := data.get("sections"):
            starts = list[int]()
            labels = list[str]()

            for s in sections:
                if isinstance(s, int):
                    starts.append(s)
                    labels.append("")
                elif isinstance(s, dict):
                    starts.append(s.get("start", 0))

                    preset = s.get("preset") or s.get("presets")
                    if isinstance(preset, list):
                        label = ", ".join(map(str, preset)) if preset else ""
                    else:
                        label = str(fallback(preset, ""))
                    labels.append(label)

            trim = data.get("trim", [[0, 0]])
            last_end = trim[0][1] if trim and isinstance(trim, list) and isinstance(trim[0], list) else starts[-1]
            ends = [*starts[1:], last_end]

            ranges = list[RangeFrame]()
            for start, end, label in zip(starts, ends, labels):
                if decimations:
                    start -= bisect_left(decimations, start)
                    end -= bisect_left(decimations, end)

                ranges.append(RangeFrame(start=start, end=end, label=label))

            scenes.append(SceneRow(color=self.get_color(), name=f"{name} (Sections)", ranges=ranges))

        # Bookmarks
        if bookmarks := data.get("user interface", {}).get("bookmarks"):
            bookmark_ranges = list[RangeFrame]()

            for b in bookmarks:
                if not isinstance(b, dict) or "frame" not in b:
                    continue

                frame = int(b["frame"])
                label = str(b.get("description", ""))

                if decimations:
                    frame -= bisect_left(decimations, frame)

                bookmark_ranges.append(RangeFrame(start=frame, label=label))

            if bookmark_ranges:
                scenes.append(SceneRow(color=self.get_color(), name=f"{name} (Bookmarks)", ranges=bookmark_ranges))

        if not scenes:
            raise ValueError(f"Could not find any sections nor bookmarks in Wobble file {name}")

        return scenes


class PythonListFramesParser(Parser):
    filter = Parser.FileFilter("Python List (Frames)", "txt")

    def parse(self, io: BinaryIO, name: str, fps: Fraction) -> SceneRow:
        with borrowed_text_wrapper(io) as wrapper:
            text = wrapper.read().strip()

        if not text:
            raise ValueError("Empty file")

        try:
            data = ast.literal_eval(text)
        except Exception as e:
            raise ValueError(f"Could not parse Python List (Frames) file: {name}") from e

        ranges = list[RangeFrame]()

        for item in data:
            try:
                ranges.append(RangeFrame.model_validate(item))
            except Exception:
                logger.warning("Skipping invalid item in %s: %s", name, item)

        if not ranges:
            raise ValueError("Empty file")

        return SceneRow(color=self.get_color(), name=name, ranges=ranges)


class PythonListTimestampsParser(Parser):
    filter = Parser.FileFilter("Python List (Timestamps)", "txt")

    def parse(self, io: BinaryIO, name: str, fps: Fraction) -> SceneRow:
        with borrowed_text_wrapper(io) as wrapper:
            text = wrapper.read().strip()

        if not text:
            raise ValueError("Empty file")

        try:
            data = ast.literal_eval(text)
        except Exception as e:
            raise ValueError(f"Could not parse Python List (Timestamps) file: {name}") from e

        ranges = list[RangeTime]()

        for item in data:
            try:
                ranges.append(RangeTime.model_validate(item))
            except Exception:
                logger.warning("Error parsing item in %s: %s", name, item)

        if not ranges:
            raise ValueError("Empty file")

        return SceneRow(color=self.get_color(), name=name, ranges=ranges)


internal_parsers: list[Parser] = [
    AssParser(),
    OGMParser(),
    MatroskaXMLParser(),
    XvidLogParser(),
    QPFileParser(),
    WobblyParser(),
    PythonListFramesParser(),
    PythonListTimestampsParser(),
]

# "Wobbly Sections (*.txt)"
# "VSEdit Bookmarks (*.bookmarks)"



from __future__ import annotations

from typing import TYPE_CHECKING

import pluggy

if TYPE_CHECKING:
    from .categories import CategoryMatcher, IterCategoryMatcher
    from .formatters import FormatterProperty, IterFormatter

hookspec = pluggy.HookspecMarker("vsview.frameprops")
hookimpl = pluggy.HookimplMarker("vsview.frameprops")


@hookspec
def vsview_frameprops_register_category_matchers() -> CategoryMatcher | IterCategoryMatcher:
    raise NotImplementedError


@hookspec
def vsview_frameprops_register_formatter_properties() -> FormatterProperty | IterFormatter:
    raise NotImplementedError
